"""Gstat
    Summarize the status of the jobs (wrapper around "squeue") using (some of) the following fields:

        +--------------+------------------------------------------------+
        | Header       | Description                                    |
        +==============+================================================+
        | "JobID"      | Job-id                                         |
        +--------------+------------------------------------------------+
        | "User"       | Username                                       |
        +--------------+------------------------------------------------+
        | "Account"    | Account name                                   |
        +--------------+------------------------------------------------+
        | "Name"       | Job name                                       |
        +--------------+------------------------------------------------+
        | "Tstart"     | Time as which the job will start / has started |
        +--------------+------------------------------------------------+
        | "Tleft"      | Maximum duration left                          |
        +--------------+------------------------------------------------+
        | "#node"      | Number of nodes claimed                        |
        +--------------+------------------------------------------------+
        | "#CPU"       | Number of CPUs claimed                         |
        +--------------+------------------------------------------------+
        | "MEM"        | Memory claimed                                 |
        +--------------+------------------------------------------------+
        | "ST"         | Status                                         |
        +--------------+------------------------------------------------+
        | "Partition"  | Partition                                      |
        +--------------+------------------------------------------------+
        | "Host"       | Hostname                                       |
        +--------------+------------------------------------------------+
        | "Dependency" | Dependency / dependencies                      |
        +--------------+------------------------------------------------+
        | "WorkDir"    | Working directory                              |
        +--------------+------------------------------------------------+

Usage:
    Gstat [options]
    Gstat [options] <JobId>...

Options:
    -U
        Limit output to the current user.

    -u, --user=<NAME>
        Limit output to user(s).
        Option may be repeated. Search by regex.

    -j, --jobid=<NAME>
        Limit output to job-id(s).
        Option may be repeated. Search by regex.

    --root=<NAME>
        Filter jobs whose workdir has this root.

    --host=<NAME>
        Limit output to host(s).
        Option may be repeated. Search by regex.

    -a, --account=<NAME>
        Limit output to account(s).
        Option may be repeated. Search by regex.

    -n, --name=<NAME>
        Limit output to job-name(s).
        Option may be repeated. Search by regex.

    -w, --workdir=<NAME>
        Limit output to job-name(s).
        Option may be repeated. Search by regex.

    --status=<NAME>
        Limit output to status.
        Option may be repeated. Search by regex.

    -p, --partition=<NAME>
        Limit output to partition(s).
        Option may be repeated. Search by regex.

    -s, --sort=<NAME>
        Sort by field.
        Option may be repeated. See description for header names.

    -r, --reverse
        Reverse sort.

    -o, --output=<NAME>
        Select output columns.
        Option may be repeated. See description for header names.

    -e, --extra=<NAME>
        Add columns.
        Option may be repeated. See description for header names.

    --full-name
        Show full user names.

    -S, --summary
        Print only summary.

    --no-header
        Suppress header.

    --no-truncate
        Print full columns, do not truncate based on terminal width.

    --width=<N>
        Set line-width (otherwise taken as terminal width).

    --colors=<NAME>
        Select color scheme from: "none", "dark". [default: "dark"]

    -l, --list
        Print selected column as list.

    -J, --joblist
        Print selected job-id(s) as list. Sort for ``Gstat -o jobid -l``.

    --abspath
        Print directories as absolute directories (default: automatic, based on distance).

    --relpath
        Print directories as relative directories (default: automatic, based on distance).

    --sep=<NAME>
        Set column separator. [default: " "]

    --long
        Print full information (each column is printed as a line).

    --debug=<FILE>
        Debug: read ``squeue -o "%all"`` from file.

    -d, --print-dependency
        Print the selected jobs as ``-d <jobid> -d <jobid> ...``.
        Use to for example ``Gsub *slurm `Gstat -d -U --partition "serial"```.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseSLURM
"""
import argparse
import os
import pwd
import re

import numpy as np

from .. import rich
from .. import squeue
from .. import table
from .. import version


class Gstat:
    def __init__(self):
        pass

    def parse_cli_args(self, cli_args: list[str] = None):
        """
        Parse command-line arguments.
        Stores the arguments as ``self.args``

        :param cli_args: Specify command-line arguments. Default: ``sys.argv[1:]``.
        """

        class Parser(argparse.ArgumentParser):
            def print_help(self):
                print(__doc__)

        parser = Parser()
        parser.add_argument("-U", action="store_true")
        parser.add_argument("-u", "--user", type=str, action="append", default=[])
        parser.add_argument("-j", "--jobid", type=str, action="append", default=[])
        parser.add_argument("--host", type=str, action="append")
        parser.add_argument("--root", type=str)
        parser.add_argument("-a", "--account", type=str, action="append")
        parser.add_argument("-n", "--name", type=str, action="append")
        parser.add_argument("-w", "--workdir", type=str, action="append")
        parser.add_argument("--status", type=str, action="append")
        parser.add_argument("-p", "--partition", type=str, action="append")
        parser.add_argument("-s", "--sort", type=str, action="append")
        parser.add_argument("-r", "--reverse", action="store_true")
        parser.add_argument("-o", "--output", type=str, action="append")
        parser.add_argument("-e", "--extra", type=str, action="append")
        parser.add_argument("--full-name", action="store_true")
        parser.add_argument("-S", "--summary", action="store_true")
        parser.add_argument("--no-header", action="store_true")
        parser.add_argument("--no-truncate", action="store_true")
        parser.add_argument("--width", type=int)
        parser.add_argument("--colors", type=str, default="dark")
        parser.add_argument("-l", "--list", action="store_true")
        parser.add_argument("-J", "--joblist", action="store_true")
        parser.add_argument("--abspath", action="store_true")
        parser.add_argument("--relpath", action="store_true")
        parser.add_argument("--sep", type=str, default=" ")
        parser.add_argument("--long", action="store_true")
        parser.add_argument("--debug", type=str)
        parser.add_argument("-d", "--print-dependency", action="store_true")
        parser.add_argument("--version", action="version", version=version)
        parser.add_argument("jobs", type=int, nargs="*")

        if cli_args is None:
            args = vars(parser.parse_args())
        else:
            args = vars(parser.parse_args(cli_args))

        if args["U"]:
            args["user"] += [pwd.getpwuid(os.getuid())[0]]

        if args["joblist"]:
            args["output"] = ["JOBID"]
            args["list"] = True

        args["jobid"] += [f"^{i:d}$" for i in args["jobs"]]

        if args["root"]:
            if args["extra"] is None:
                args["extra"] = ["WorkDir"]
            elif "WorkDir" not in args["extra"]:
                args["extra"] += ["WorkDir"]

        # store for later use
        self.args = args

    def read(self):
        """
        Read from queuing system.
        Stores data on ``self.lines``
        Store print info as``self.columns``, ``self.header``, ``self.alias``, ``self.aliasInv``.
        """

        # -- field-names and print settings --

        # conversion map: default field-names -> custom field-names
        alias = {
            "JOBID": "JobID",
            "USER": "User",
            "ACCOUNT": "Account",
            "NAME": "Name",
            "START_TIME": "Tstart",
            "TIME_LEFT": "Tleft",
            "NODES": "#node",
            "CPUS": "#CPU",
            "CPUS_R": "#CPU(R)",
            "CPUS_PD": "#CPU(PD)",
            "MIN_MEMORY": "MEM",
            "ST": "ST",
            "NODELIST(REASON)": "Host",
            "PARTITION": "Partition",
            "DEPENDENCY": "Dependency",
            "WORK_DIR": "WorkDir",
        }

        # conversion map: custom field-names -> default field-names
        aliasInv = {alias[key].upper(): key for key in alias}

        # rename command line options -> default field-names
        # - add key-names
        aliasInv["STATUS"] = "ST"
        # - apply conversion
        for key in [key for key in self.args]:
            if key.upper() in aliasInv:
                self.args[aliasInv[key.upper()]] = self.args.pop(key)

        # print settings of all columns
        # - "width"   : minimum width, adapted to print width (min_width <= width <= real_width)
        # - "align"   : alignment of the columns (except the header)
        # - "priority": priority of column expansing, columns marked "True" are expanded first
        prio = {"priority": True, "default": True}
        noprio = {"priority": False, "default": True}
        nodefault = {"priority": True, "default": False}
        columns = [
            {"key": "JOBID", "width": 7, "align": ">", **prio},
            {"key": "USER", "width": 7, "align": "<", **prio},
            {"key": "ACCOUNT", "width": 7, "align": "<", **prio},
            {"key": "NAME", "width": 11, "align": "<", **noprio},
            {"key": "START_TIME", "width": 6, "align": ">", **prio},
            {"key": "TIME_LEFT", "width": 5, "align": ">", **prio},
            {"key": "NODES", "width": 5, "align": ">", **prio},
            {"key": "CPUS", "width": 4, "align": ">", **prio},
            {"key": "MIN_MEMORY", "width": 3, "align": ">", **prio},
            {"key": "ST", "width": 2, "align": "<", **prio},
            {"key": "PARTITION", "width": 9, "align": "<", **noprio},
            {"key": "NODELIST(REASON)", "width": 5, "align": "<", **noprio},
            {"key": "DEPENDENCY", "width": 5, "align": "<", **nodefault},
            {"key": "WORK_DIR", "width": 7, "align": "<", **nodefault},
        ]

        # header
        header = {
            column["key"]: rich.String(alias[column["key"]], align=column["align"])
            for column in columns
        }

        if self.args["root"]:
            header["WORK_DIR"] = rich.String(header["WORK_DIR"].data + " " + self.args["root"])

        # select color theme
        theme = squeue.colors(self.args["colors"].lower())

        # -- load the output of "squeue" --

        if not self.args["debug"]:

            lines = squeue.read_interpret(theme=theme)

        else:

            lines = squeue.read_interpret(
                data=open(self.args["debug"]).read(),
                now=os.path.getctime(self.args["debug"]),
                theme=theme,
            )

        # -- convert paths ---

        if self.args["root"]:
            root = self.args["root"]
            lines = [
                i for i in lines if not os.path.relpath(i["WORK_DIR"].data, root).startswith("..")
            ]
            for line in lines:
                line["WORK_DIR"].data = os.path.relpath(line["WORK_DIR"].data, root)
        elif self.args["abspath"]:
            for line in lines:
                line["WORK_DIR"].data = os.path.abspath(line["WORK_DIR"].data)
        elif self.args["relpath"]:
            for line in lines:
                line["WORK_DIR"].data = os.path.relpath(line["WORK_DIR"].data)
        else:
            for line in lines:
                if len(os.path.relpath(line["WORK_DIR"].data).split("../")) < 3:
                    line["WORK_DIR"].data = os.path.relpath(line["WORK_DIR"].data)

        # -- limit based on command-line options --

        for key in [
            "USER",
            "ACCOUNT",
            "NAME",
            "JOBID",
            "ST",
            "NODELIST(REASON)",
            "PARTITION",
            "WORK_DIR",
        ]:

            if not self.args[key]:
                continue

            # limit data
            lines = [
                line
                for line in lines
                if sum(1 if re.match(n, str(line[key])) else 0 for n in self.args[key])
            ]

            # color-highlight selected columns
            # - apply to all remaining lines
            for line in lines:
                line[key].color = theme["selection"]
            # - apply to the header
            header[key].color = theme["selection"]

        if self.args["root"]:
            key = "WORK_DIR"

            for line in lines:
                line[key].color = theme["selection"]

            header[key].color = theme["selection"]

        # -- sort --

        if self.args["sort"]:
            sortkeys = [aliasInv[key.upper()] for key in self.args["sort"]]
            reversed = False
        else:
            sortkeys = ["JOBID", "PARTITION"]
            reversed = False

        if self.args["reverse"]:
            reversed = not reversed

        idx = np.lexsort([[i[key] for i in lines] for key in sortkeys])
        if reversed:
            idx = idx[::-1]
        lines = [lines[i] for i in idx]

        # -- select columns --

        if self.args["extra"]:
            keys = [aliasInv[key.upper()] for key in self.args["extra"]]
            extra = [column for column in columns if column["key"] in keys]
        else:
            extra = []

        if self.args["output"]:
            keys = [aliasInv[key.upper()] for key in self.args["output"]]
            columns = [column for column in columns if column["key"] in keys]
        else:
            columns = [column for column in columns if column["default"]]

        columns += extra

        # store for later use
        self.lines = lines
        self.columns = columns
        self.header = header
        self.alias = alias
        self.aliasInv = aliasInv

    def print_all(self):
        """
        Normal print
        """

        # print all fields and quit
        if self.args["long"]:

            table.print_long(self.lines)
            return

        # print as list and quit
        if self.args["list"]:

            if len(self.columns) > 1:
                raise OSError("Error: Only one field can be selected")

            table.print_list(self.lines, self.columns[0]["key"], self.args["sep"])

        # print columns
        table.print_columns(
            self.lines,
            self.columns,
            self.header,
            self.args["no_truncate"],
            self.args["sep"],
            self.args["width"],
            not self.args["no_header"],
        )

    def print_summary(self):
        """
        Print summary.
        """

        # print settings for the summary
        columns_summary = [
            {"key": "USER", "width": 7, "align": "<", "priority": True},
            {"key": "ACCOUNT", "width": 7, "align": "<", "priority": False},
            {"key": "CPUS", "width": 4, "align": ">", "priority": True},
            {"key": "CPUS_R", "width": 6, "align": ">", "priority": True},
            {"key": "CPUS_PD", "width": 6, "align": ">", "priority": True},
            {"key": "PARTITION", "width": 9, "align": "<", "priority": False},
        ]

        # header
        header_summary = {
            column["key"]: rich.String(self.alias[column["key"]], align=column["align"])
            for column in columns_summary
        }

        # -- summarize information --

        # get names of the different users
        users = sorted({str(line["USER"]) for line in self.lines})

        # start a new list of "user information", summed on the relevant users
        users = [{"USER": rich.String(key)} for key in users]

        # loop over users
        for user in users:

            # - isolate jobs for this user
            N = [line for line in self.lines if str(line["USER"]) == str(user["USER"])]

            # - get (a list of) partition(s)/account(s)
            user["PARTITION"] = rich.String(",".join(list({str(line["PARTITION"]) for line in N})))
            user["ACCOUNT"] = rich.String(",".join(list({str(line["ACCOUNT"]) for line in N})))

            # - count used CPU (per category)
            user["CPUS"] = rich.Integer(sum(int(line["CPUS"]) for line in N))
            user["CPUS_R"] = rich.Integer(sum(int(line["CPUS_R"]) for line in N))
            user["CPUS_PD"] = rich.Integer(sum(int(line["CPUS_PD"]) for line in N))

            # - remove zeros from output for more intuitive output
            for key in ["CPUS_R", "CPUS_PD"]:
                if int(user[key]) == 0:
                    user[key] = rich.Integer("-")

        # rename field
        lines = users

        # -- sort --

        # default sort
        lines.sort(key=lambda line: line["USER"], reverse=self.args["reverse"])

        # optional: sort by key(s)
        if self.args["sort"]:

            # get available keys in the setting with fewer columns
            keys = [self.alias[column["key"]].upper() for column in columns_summary]

            # filter sort keys that are not available in this mode
            self.args["sort"] = [key for key in self.args["sort"] if key.upper() in keys]

            # apply sort
            for key in self.args["sort"]:
                lines.sort(
                    key=lambda line: line[self.aliasInv[key.upper()]], reverse=self.args["reverse"]
                )

        # -- print --

        table.print_columns(
            lines,
            columns_summary,
            header_summary,
            self.args["no_truncate"],
            self.args["sep"],
            self.args["width"],
            not self.args["no_header"],
        )

    def print(self):
        """
        Print.
        """

        if self.args["print_dependency"]:
            if len(self.lines) > 0:
                print("-d " + " -d ".join([str(line["JOBID"]) for line in self.lines]))
        elif not self.args["summary"]:
            self.print_all()
        else:
            self.print_summary()


def main():
    try:
        p = Gstat()
        p.parse_cli_args()
        p.read()
        p.print()
    except Exception as e:
        print(e)
        return 1



import re
import subprocess

from . import duration
from . import fileio
from . import files
from . import memory
from . import ps
from . import rich
from . import sacct
from . import scripts
from . import sinfo
from . import squeue
from . import table
from ._version import version
from ._version import version_tuple


def sbatch(options, verbose=False, dry_run=False):
    """
    Submit job and return the job-id.
    """

    assert type(options) == list

    if dry_run or verbose:
        print(" ".join(["sbatch"] + options))

    if dry_run:
        return None

    out = subprocess.check_output(["sbatch"] + options).decode("utf-8")

    if verbose:
        print(out, end="")

    return int(re.split(r"(Submitted batch job)([\ ]*)([0-9]*)(.*)", out)[3])



project = "GooseSLURM"
copyright = "2018-2022, Tom de Geus"
author = "Tom de Geus"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinxarg.ext",
]

html_theme = "furo"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]



import GooseSLURM as gs

# job-options
sbatch = {
    "job-name": "parallel",
    "out": "job.slurm.out",
    "nodes": 1,
    "ntasks": 1,
    "cpus-per-task": 5,
    "partition": "debug",
}

# command to run
command = "parallel --max-procs=${SLURM_CPUS_PER_TASK} 'echo {1} {2}' ::: {1..12} ::: summer winter"

# write SLURM file
open("job.slurm", "w").write(gs.scripts.plain(command=command, **sbatch))



r"""Gsub
    Submit job-scripts and add the "--chdir" option
    to run the scripts from the directory in with the sbatch-file is stored.

Usage:
    Gsub [options] <files>...

Arguments:
    Job-scripts.

Options:
    --dry-run
        Print commands to screen, without executing.

    --verbose
        Verbose all commands and their output.

    --log = FILENAME
        Log the JobIDs to a YAML-file (updated after each submit).
        Existing log files are appended.

    --delay = FLOAT
        Seconds to wait between submitting jobs. [default: 0.1]

    -r, --repeat = INT
        Submit using dependencies such that the job will be repeated 'n' times. [default: 1]

    --serial
        Submit using dependencies such that jobs are run after each other.

    -d, --dependency = ARG (sbatch option)
        Defer the start of this job until the specified dependencies have been satisfied completed.

    -w, --wait (sbatch option)
        Do not exit until the submitted job terminates.

    -c, --constraint = ARG (sbatch option)
        Nodes can have features assigned to them by the Slurm administrator.

    -Q, --quiet
        Do no show progress-bar.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseSLURM
"""
import argparse
import os
import re
import subprocess
import time

import tqdm

from .. import fileio
from .. import version


def sbatch(options, verbose=False, dry_run=False):
    """
    Submit job and return the job-id.
    """

    assert type(options) == list

    if dry_run or verbose:
        print(" ".join(["sbatch"] + options))

    if dry_run:
        return None

    out = subprocess.check_output(["sbatch"] + options).decode("utf-8")

    if verbose:
        print(out, end="")

    return int(re.split(r"(Submitted batch job)([\ ]*)([0-9]*)(.*)", out)[3])


def read_log(files, logfile=None):
    """
    Read existing log-file.
    If the file does not exists an empty log is returned.

    :param files: List of files to submit.
    :param logfile: Filename of the log-file.
    :return: Log as dict.
    """

    log = {filename: [] for filename in files}

    if logfile is None:
        return log

    if not os.path.isfile(os.path.realpath(logfile)):
        return log

    log = fileio.YamlRead(logfile)

    if type(log) != dict:
        raise OSError("Unable to interpret log file")

    for filename in files:
        if filename not in log:
            raise OSError(f'"{filename} not in log file"')
        if type(log[filename]) != list:
            raise OSError(f'Log of "{filename}" not interpretable')

    return log


def main():
    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("--dry-run", action="store_true")
    parser.add_argument("--verbose", action="store_true")
    parser.add_argument("-l", "--log", type=str)
    parser.add_argument("--delay", type=float, default=0.1)
    parser.add_argument("-r", "--repeat", type=int, default=1)
    parser.add_argument("-d", "--dependency", type=str)
    parser.add_argument("-w", "--wait", action="store_true")
    parser.add_argument("-c", "--constraint", type=str)
    parser.add_argument("-Q", "--quiet", action="store_true")
    parser.add_argument("--serial", action="store_true")
    parser.add_argument("files", nargs="*", type=str)
    args = parser.parse_args()
    dargs = vars(args)

    log = read_log(args.files, args.log)
    jobid = ""

    for filename in args.files:
        if not os.path.isfile(filename):
            raise OSError(f'"{filename}" does not exist')

    pbar = tqdm.tqdm(args.files, disable=args.quiet)

    for ifile, file in enumerate(pbar):
        for rep in range(int(args.repeat)):
            pbar.set_description(file)
            path, name = os.path.split(file)
            options = ["--chdir", os.path.abspath(path)]
            if args.wait:
                options += ["--wait"]
            for opt in ["constraint", "dependency"]:
                if dargs[opt]:
                    options += [f"--{opt:s}", dargs[opt]]
            if rep or (args.serial and ifile):
                options += ["--dependency", str(jobid)]
            options += [name]
            jobid = sbatch(options, verbose=args.verbose, dry_run=args.dry_run)
            log[file] += [jobid]
            if args.log:
                fileio.YamlDump(args.log, log)
            time.sleep(float(args.delay))



import os
import subprocess
import unittest

import dummyslurm
import GooseSLURM


class Test_Gdel(unittest.TestCase):
    """
    Test Gdel commands.
    """

    def test_basic(self):

        myjob = "myjob.slurm"

        for filename in [dummyslurm.logfile, myjob]:
            if os.path.isfile(filename):
                os.remove(filename)

        with open(myjob, "w") as file:
            file.write(GooseSLURM.scripts.plain(myjob))

        subprocess.check_output(["Gsub", "--quiet", "--repeat", "4", myjob])
        subprocess.run(["Gdel"], capture_output=True, text=True, input="Y")

        log = GooseSLURM.fileio.YamlRead(dummyslurm.logfile)

        self.assertEqual(len(log), 0)

        os.remove(dummyslurm.logfile)
        os.remove(myjob)

    def test_name(self):

        myjobs = ["myjob1.slurm", "myjob2.slurm"]

        for filename in [dummyslurm.logfile] + myjobs:
            if os.path.isfile(filename):
                os.remove(filename)

        for myjob in myjobs:
            with open(myjob, "w") as file:
                file.write(GooseSLURM.scripts.plain(myjob))

        for myjob in myjobs:
            subprocess.run(["Gsub", "--quiet", myjob])

        subprocess.run(["Gdel", "-n", f"^{myjobs[0]}$"], capture_output=True, text=True, input="Y")

        log = GooseSLURM.fileio.YamlRead(dummyslurm.logfile)

        self.assertEqual(len(log), 1)
        self.assertEqual(log[0]["job_name"], myjobs[1])

        os.remove(dummyslurm.logfile)

        for myjob in myjobs:
            os.remove(myjob)


if __name__ == "__main__":

    unittest.main()



import GooseSLURM as gs

# job-options
sbatch = {
    "job-name": "tempdir",
    "out": "job.slurm.out",
    "nodes": 1,
    "ntasks": 1,
    "cpus-per-task": 1,
    "partition": "debug",
}

# command to run
command = """
# simplest example in the world, sleep a bit to allow a bit of monitoring
echo "hello world" > "test.log"
sleep 10
"""

# write SLURM file
open("job.slurm", "w").write(gs.scripts.tempdir(command=command, **sbatch))



from . import duration
from . import memory


def plain(
    command=[],
    shell="#!/bin/bash -l",
    **sbatch,
):
    r"""
    Return simple SBATCH-file (as text).

    :options:

        **command** (``<str>`` | ``<list>``)
            Command(s) to execute.
            If the input is a list each entry is included as an individual line.

        **shell** (``<str>``)
            The shell to use.

    :SBATCH options:

        **mem** (``<int>`` | ``<str>``)
            Memory claim (may be human readable, see ``GooseSLUM.memory.asSlurm``).

        **time**  (``<str>``)
            Wall-time claim (may be human readable, see ``GooseSLUM.duration.asSlurm``).

        **out** (``<str``>)
            Name of the output file, e.g. ``myjob_%j.out``.

        ...
    """

    # convert to string (if needed)
    if not isinstance(command, str):
        command = "\n".join(command)

    # convert sbatch options
    for key, item in sbatch.items():
        if key in ["time"]:
            sbatch[key] = duration.asSlurm(item)
        if key in ["mem"]:
            sbatch[key] = memory.asSlurm(item)

    # - convert to string
    sbatch = "\n".join([f"#SBATCH --{key:s} {str(arg):s}" for key, arg in sbatch.items()])

    return """{shell:s}
{sbatch:s}

{command:s}
  """.format(
        shell=shell,
        sbatch=sbatch,
        command=command,
    )


def tempdir(remove=[], command=[], shell="#!/bin/bash -l", **sbatch):
    r"""
    Return SBATCH-file (as text) that uses a temporary working directory on the compute node.

    :options:

        **remove** (``<list>``)
            List with files/folders to remove from the temporary directory before copying.

        **command** (``<str>`` | ``<list>``)
            Command(s) to execute.
            If the input is a list each entry is included as an individual line.

        **shell** (``<str>``)
            The shell to use.

    :SBATCH options:

        **mem** (``<int>`` | ``<str>``)
            Memory claim (may be human readable, see ``GooseSLUM.memory.asSlurm``).

        **time**  (``<str>``)
            Wall-time claim (may be human readable, see ``GooseSLUM.duration.asSlurm``).

        **out** (``<str``>)
            Name of the output file, e.g. ``myjob_%j.out``.

        ...
    """

    # convert to string
    if len(remove) > 0:
        remove = "rm -r " + " ".join(remove)
    else:
        remove = "# rm -r ..."

    # convert to string (if needed)
    if not isinstance(command, str):
        command = "\n".join(command)

    # convert sbatch options
    for key, item in sbatch.items():
        if key in ["time"]:
            sbatch[key] = duration.asSlurm(item)
        if key in ["mem"]:
            sbatch[key] = memory.asSlurm(item)

    # extract name of the output file
    outfile = sbatch["out"]

    # - convert to string
    sbatch = "\n".join([f"#SBATCH --{key:s} {str(arg):s}" for key, arg in sbatch.items()])

    return """{shell:s}
{sbatch:s}

# I. Define directory names [DO NOT CHANGE]
# =========================================

# get name of the temporary directory working directory, physically on the compute-node
workdir="${{TMPDIR}}"

# get current directory
# (every file/folder below this directory is copied to the compute node)
submitdir="$(pwd)"

# 1. Transfer to node [DO NOT CHANGE]
# ===================================

# create/empty the temporary directory on the compute node
if [ ! -d "${{workdir}}" ]; then
  mkdir -p "${{workdir}}"
else
  rm -rf "${{workdir}}"/*
fi

# change current directory to the location of the sbatch command
# ("submitdir" is somewhere in the home directory on the head node)
cd "${{submitdir}}"
# copy all files/folders in "submitdir" to "workdir"
# ("workdir" == temporary directory on the compute node)
cp -prf * ${{workdir}}
# change directory to the temporary directory on the compute-node
cd ${{workdir}}

# 3. Function to transfer back to the head node [DO NOT CHANGE]
# =============================================================

# define clean-up function
function clean_up {{
  # - remove log-file on the compute-node, to avoid the one created by SLURM is overwritten
  rm {outfile:s}
  # - delete temporary files from the compute-node, before copying
  {remove:s}
  # - change directory to the location of the sbatch command (on the head node)
  cd "${{submitdir}}"
  # - copy everything from the temporary directory on the compute-node
  cp -prf "${{workdir}}"/* .
  # - erase the temporary directory from the compute-node
  rm -rf "${{workdir}}"/*
  rm -rf "${{workdir}}"
  # - exit the script
  exit
}}

# call "clean_up" function when this script exits, it is run even if SLURM cancels the job
trap 'clean_up' EXIT

# 2. Execute [MODIFY COMPLETELY TO YOUR NEEDS]
# ============================================

{command:s}
  """.format(
        shell=shell,
        sbatch=sbatch,
        command=command,
        outfile=outfile,
        remove=remove,
    )



import argparse
import inspect
import os
import pwd
import re
import sys
import textwrap

import yaml

version = "1.2.3"

dirname = os.path.dirname(os.path.abspath(__file__))
logfile = "_sbatch.yaml"


def sbatch():
    """
    Dummy ``sbatch`` command.
    A local file '_sbatch.yaml' in the current working directory will contain the history of
    'submitted' jobs.
    """

    log = []

    if os.path.isfile(os.path.realpath(logfile)):
        with open(logfile) as file:
            log = yaml.load(file.read(), Loader=yaml.FullLoader)

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    funcname = inspect.getframeinfo(inspect.currentframe()).function
    doc = textwrap.dedent(inspect.getdoc(globals()[funcname]))

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=doc)
    parser.add_argument("script", type=str)
    parser.add_argument("-A", "--account", type=str, default="default")
    parser.add_argument("-c", "--cpus-per-task", type=str, default="1")
    parser.add_argument("-d", "--dependency", type=int)
    parser.add_argument("-J", "--job-name", type=str)
    parser.add_argument("-N", "--nodes", type=str, default="1")
    parser.add_argument("-n", "--ntasks", type=str, default="1")
    parser.add_argument("-p", "--partition", type=str, default="serial")
    parser.add_argument("-t", "--time", type=str, default="1:00:00")
    parser.add_argument("--mem", type=str, default="5000000000")
    parser.add_argument("--chdir", type=str)
    args = parser.parse_args()

    if args.job_name is None:
        args.job_name = args.script

    # read SBATCH options from script
    if os.path.isfile(args.script):
        extra = []
        with open(args.script) as file:
            lines = file.read().split("\n")
        for line in lines:
            if re.match("^#SBATCH", line):
                extra += re.split("^#SBATCH", line)[1].strip().rstrip().split(" ")
        if len(extra) > 0:
            args = parser.parse_args(extra + sys.argv[1:])

    jobid = len(log) + 1
    mylog = vars(args).copy()
    for key, value in dict(mylog).items():
        if value is None:
            del mylog[key]

    mylog["state"] = "PD"
    mylog["jobid"] = jobid
    mylog["user"] = pwd.getpwuid(os.getuid())[0]
    mylog["command"] = os.path.abspath(mylog["script"])
    if args.chdir:
        mylog["workdir"] = args.chdir
    else:
        mylog["workdir"] = os.path.dirname(os.path.abspath(mylog["script"]))

    log += [mylog]

    print(f"Submitted batch job {jobid:d}")

    with open(logfile, "w") as file:
        yaml.dump(log, file)


def scancel():
    """
    Dummy ``scancel`` command.
    A local file '_sbatch.yaml' in the current working directory will contain the history of
    'submitted' jobs.
    """

    log = []

    if os.path.isfile(os.path.realpath(logfile)):
        with open(logfile) as file:
            log = yaml.load(file.read(), Loader=yaml.FullLoader)

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    funcname = inspect.getframeinfo(inspect.currentframe()).function
    doc = textwrap.dedent(inspect.getdoc(globals()[funcname]))

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=doc)
    parser.add_argument("jobid", type=int, nargs="*")
    args = parser.parse_args()

    log = [i for i in log if i["jobid"] not in args.jobid]

    with open(logfile, "w") as file:
        yaml.dump(log, file)


def squeue():
    """
    Dummy ``squeue`` command.
    Reads from local file '_sbatch.yaml' in the current working directory created using the dummy
    ``sbatch`` command.
    """

    log = []

    if os.path.isfile(os.path.realpath(logfile)):
        with open(logfile) as file:
            log = yaml.load(file.read(), Loader=yaml.FullLoader)

    if sys.argv[1:] == ["-o", "%all"]:
        keys = [
            "ACCOUNT",
            "TRES_PER_NODE",
            "MIN_CPUS",
            "MIN_TMP_DISK",
            "END_TIME",
            "FEATURES",
            "GROUP",
            "OVER_SUBSCRIBE",
            "JOBID",
            "NAME",
            "COMMENT",
            "TIME_LIMIT",
            "MIN_MEMORY",
            "REQ_NODES",
            "COMMAND",
            "PRIORITY",
            "QOS",
            "REASON",
            "ST",
            "USER",
            "RESERVATION",
            "WCKEY",
            "EXC_NODES",
            "NICE",
            "S:C:T",
            "JOBID",
            "EXEC_HOST",
            "CPUS",
            "NODES",
            "DEPENDENCY",
            "ARRAY_JOB_ID",
            "GROUP",
            "SOCKETS_PER_NODE",
            "CORES_PER_SOCKET",
            "THREADS_PER_CORE",
            "ARRAY_TASK_ID",
            "TIME_LEFT",
            "TIME",
            "NODELIST",
            "CONTIGUOUS",
            "PARTITION",
            "PRIORITY",
            "NODELIST(REASON)",
            "START_TIME",
            "STATE",
            "UID",
            "SUBMIT_TIME",
            "LICENSES",
            "CORE_SPEC",
            "SCHEDNODES",
            "WORK_DIR",
        ]

        alias = {
            "ACCOUNT": "account",
            "CPUS": "cpus_per_task",
            "DEPENDENCY": "dependency",
            "JOBID": "jobid",
            "MIN_MEMORY": "mem",
            "NAME": "job_name",
            "NODES": "nodes",
            "PARTITION": "partition",
            "ST": "state",
            "TIME": "time",
            "USER": "user",
            "WORK_DIR": "workdir",
            "HOST": "host",
            "TIME_START": "time_start",
        }

        print("|".join([i.upper() for i in keys]))

        for i in log:
            print("|".join([str(i.get(alias.get(key, "NONE"), "N/A")) for key in keys]))

        return 0

    raise OSError("Command not implemented")


def scontrol():

    log = []

    if os.path.isfile(os.path.realpath(logfile)):
        with open(logfile) as file:
            log = yaml.load(file.read(), Loader=yaml.FullLoader)

    if re.match(r"^(show job )([0-9]*)", " ".join(sys.argv[1:])):

        jobid = int(re.split(r"^(show job )([0-9]*)", " ".join(sys.argv[1:]))[2])

        keys = [
            "JobId",
            "JobName",
            "UserId",
            "GroupId",
            "MCS_label",
            "Priority",
            "Nice",
            "Account",
            "QOS",
            "JobState",
            "Reason",
            "Dependency",
            "Requeue",
            "Restarts",
            "BatchFlag",
            "Reboot",
            "ExitCode",
            "RunTime",
            "TimeLimit",
            "TimeMin",
            "SubmitTime",
            "EligibleTime",
            "AccrueTime",
            "StartTime",
            "EndTime",
            "Deadline",
            "SuspendTime",
            "SecsPreSuspend",
            "LastSchedEval",
            "Partition",
            "AllocNode:Sid",
            "ReqNodeList",
            "ExcNodeList",
            "NodeList",
            "BatchHost",
            "NumNodes",
            "NumCPUs",
            "NumTasks",
            "CPUs/Task",
            "ReqB:S:C:T",
            "TRES",
            "Socks/Node",
            "NtasksPerN:B:S:C",
            "CoreSpec",
            "MinCPUsNode",
            "MinMemoryCPU",
            "MinTmpDiskNode",
            "Features",
            "DelayBoot",
            "OverSubscribe",
            "Contiguous",
            "Licenses",
            "Network",
            "Command",
            "WorkDir",
            "StdErr",
            "StdIn",
            "StdOut",
            "Power",
        ]

        alias = {
            "Account": "account",
            "Command": "command",
            "JobId": "jobid",
            "JobName": "job_name",
            "UserId": "user",
        }

        for i in log:
            if i["jobid"] == jobid:
                ret = [f"{key}=" + str(i.get(alias.get(key, "NONE"), "")) for key in keys]
                ret = textwrap.wrap(" ".join(ret), subsequent_indent="    ", break_long_words=False)
                ret = "\n".join(ret)
                print(ret)
                return 0

        raise OSError("JobID not found")

    raise OSError("Command not implemented")


def sacct():

    args = sys.argv[1:]
    log = []

    if os.path.isfile(os.path.realpath(logfile)):
        with open(logfile) as file:
            log = yaml.load(file.read(), Loader=yaml.FullLoader)

    allocations = False
    if "-X" in sys.argv[1:]:
        allocations = True
        args.remove("-X")

    if re.match(r"^(-p -l -j )([0-9\,]*)", " ".join(args)):

        jobids = map(int, re.split(r"^(-p -l -j )([0-9\,]*)", " ".join(args))[2].split(","))

        keys = [
            "JobID",
            "JobIDRaw",
            "JobName",
            "Partition",
            "MaxVMSize",
            "MaxVMSizeNode",
            "MaxVMSizeTask",
            "AveVMSize",
            "MaxRSS",
            "MaxRSSNode",
            "MaxRSSTask",
            "AveRSS",
            "MaxPages",
            "MaxPagesNode",
            "MaxPagesTask",
            "AvePages",
            "MinCPU",
            "MinCPUNode",
            "MinCPUTask",
            "AveCPU",
            "NTasks",
            "AllocCPUS",
            "Elapsed",
            "State",
            "ExitCode",
            "AveCPUFreq",
            "ReqCPUFreqMin",
            "ReqCPUFreqMax",
            "ReqCPUFreqGov",
            "ReqMem",
            "ConsumedEnergy",
            "MaxDiskRead",
            "MaxDiskReadNode",
            "MaxDiskReadTask",
            "AveDiskRead",
            "MaxDiskWrite",
            "MaxDiskWriteNode",
            "MaxDiskWriteTask",
            "AveDiskWrite",
            "AllocGRES",
            "ReqGRES",
            "ReqTRES",
            "AllocTRES",
            "TRESUsageInAve",
            "TRESUsageInMax",
            "TRESUsageInMaxNode",
            "TRESUsageInMaxTask",
            "TRESUsageInMin",
            "TRESUsageInMinNode",
            "TRESUsageInMinTask",
            "TRESUsageInTot",
            "TRESUsageOutMax",
            "TRESUsageOutMaxNode",
            "TRESUsageOutMaxTask",
            "TRESUsageOutAve",
            "TRESUsageOutTot",
        ]

        alias = {
            "JobID": "jobid",
            "JobName": "job_name",
        }

        lines = []

        for i in log:
            if i["jobid"] in jobids:

                base = [str(i.get(alias.get(key, "NONE"), "")) for key in keys]
                batch = [r for r in base]
                extern = [r for r in base]
                batch[0] = batch[0] + ".batch"
                extern[0] = extern[0] + ".extern"
                lines.append("|".join(base) + "|")
                if not allocations:
                    lines.append("|".join(batch) + "|")
                    lines.append("|".join(extern) + "|")

        if len(lines) > 0:
            print("|".join(keys) + "|")
            print("\n".join(lines))
            return 0

        raise OSError("JobID not found")

    raise OSError("Command not implemented")



from __future__ import annotations

import argparse
import datetime
import json
import os
import pwd
import re
import subprocess
import sys
from collections import defaultdict

import numpy as np

from . import duration
from . import rich
from . import table
from ._version import version


def _read(cmd: str) -> list[dict]:
    r"""
    Read command and interpret.
    Requires ``-p`` and ``-l``.

    :param jobid: The jobid to read.
    :return: List of dictionaries, that contain the different fields. All data are strings.
    """

    data = subprocess.check_output(cmd, shell=True).decode("utf-8")

    head, data = data.split("\n", 1)
    data = list(filter(None, data.split("\n")))
    head = head.split("|")

    lines = []
    for line in data:

        info = {}

        for key, val in zip(head, line.split("|")):
            if len(key.strip()) > 0:
                info[key.strip()] = val.strip()

        lines += [info]

    return lines


def read_job(jobid: int | str) -> list[dict]:
    r"""
    Read ``sacct -p -l -j jobid`.

    :param jobid: The jobid to read.
    :return: List of dictionaries, that contain the different fields. All data are strings.
    """

    return _read(f"sacct -p -l -j {str(jobid)}")


def _asdate(text: str):

    if text[0] != "-":
        return text

    d = datetime.datetime.now() - datetime.timedelta(seconds=duration.asSeconds(text[1:]))
    return d.strftime("%Y-%m-%dT%H:%M:%S")


def cli_parser() -> argparse.ArgumentParser:
    """
    Return parser of command-line arguments.
    """

    help = """
    Display available information from ``sacct -j jobid...`` or ``sacct [OPTION]``.

    *   The time fields can be specified as rich
        (e.g. ``-S="-1h"`` for jobs that started at maximum one hour ago)
        in addition to the default format of ``sacct``.

    *   As state use: running / r, completed / cd, failed / f, timeout / to,
        resizing / rs, deadline / dl, node_fail / nf.

    *   The output can be returned in JSON format (``--json``).

    *   Extra columns can be added (``--extra``), see ``sacct --helpformat``.
        Commonly used are ``--extra="WorkDir"``.

    *   The following column abbreviations are used (depending on the output):
        -  ``State``: ``ST``
        -  ``ExitCode``: ``exit``

    .. tip::

        Use ``seff jobid`` to get more information, such as memory usage.
    """

    parser = argparse.ArgumentParser(
        description=help,
    )

    append = dict(type=str, action="append", default=[])
    parser.add_argument("-X", "--allocations", action="store_true", help="Include only main job.")
    parser.add_argument("-j", "--json", action="store_true", help="Print in JSON format.")
    parser.add_argument("--sep", type=str, default=" ", help="Column separator.")
    parser.add_argument("--no-truncate", action="store_true", help="Print without fitting screen.")
    parser.add_argument("--sort", help="Sort based on column.", **append)
    parser.add_argument("--reverse", action="store_true", help="Reverse order.")
    parser.add_argument("--infer", type=str, help="Read extra data from ``JOBID.infer``.")
    parser.add_argument("-e", "--extra", help="Extra columns.", **append)
    parser.add_argument(
        "--abspath", action="store_true", help="Print directories as absolute (default: automatic)."
    )
    parser.add_argument(
        "--relpath", action="store_true", help="Print directories as relative (default: automatic)."
    )
    parser.add_argument(
        "--root",
        type=str,
        help="Filter jobs whose WorkDir has this root (WorkDir printed relative to this root).",
    )
    parser.add_argument(
        "-L", "--allclusters", action="store_true", help="Display jobs ran on all clusters."
    )
    parser.add_argument(
        "-a", "--allusers", action="store_true", help="All users (default: only current user)."
    )
    parser.add_argument("-T", "--truncate", action="store_true", help="Truncate time.")
    parser.add_argument("-S", "--starttime", type=str, help="Job started after time.")
    parser.add_argument("-E", "--endtime", type=str, help="Job end before time.")
    parser.add_argument("-i", "--nnodes", type=str, help="Jobs which ran on this many nodes.")
    parser.add_argument("-I", "--ncpus", type=str, help="Jobs which ran on this many cpus.")
    parser.add_argument(
        "-k", "--timelimit-min", type=str, help="Only send data about jobs with this timelimit."
    )
    parser.add_argument(
        "-K", "--timelimit-max", type=str, help="Only send data about jobs with this timelimit."
    )
    parser.add_argument("-r", "--partition", help="(Comma separated list of) partitions.", **append)
    parser.add_argument("-s", "--state", help="Select state(s).", **append)
    parser.add_argument("-N", "--nodelist", help="Select nodelist(s).", **append)
    parser.add_argument("-M", "--clusters", help="Select cluster(s).", **append)
    parser.add_argument("-A", "--account", help="Select account(s).", **append)
    parser.add_argument("-u", "--user", help="Select username(s).", **append)
    parser.add_argument("--uid", help="Select user-id(s).", **append)
    parser.add_argument("-U", action="store_true", help="Select current user.")
    parser.add_argument("-g", "--group", help="Select group(s).", **append)
    parser.add_argument("--gid", help="Select group-id(s).", **append)
    parser.add_argument("--name", help="Select job-name(s).", **append)
    parser.add_argument("-q", "--qos", help="Select qos(s).", **append)
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("jobid", type=int, nargs="*", help="JobID(s) to read.")
    return parser


def Gacct(args: list[str]):
    """
    Command-line tool to print datasets from a file, see ``--help``.
    :param args: Command-line arguments (should be all strings).
    """

    parser = cli_parser()
    args = parser.parse_args(args)
    extra = [i for i in args.extra]

    if args.root is not None:
        if "WorkDir" not in extra:
            extra += ["WorkDir"]

    opts = ["-p", "-l"]
    if args.allocations:
        opts += ["-X"]
    if args.allusers:
        opts += ["-a"]
    if args.allclusters:
        opts += ["-A"]
    if args.truncate:
        opts += ["-T"]
    if args.starttime:
        opts += ["-S", _asdate(args.starttime)]
    if args.endtime:
        opts += ["-E", _asdate(args.endtime)]
    if args.nnodes:
        opts += ["-i", args.nnodes]
    if args.ncpus:
        opts += ["-I", args.ncpus]
    if args.timelimit_min:
        opts += ["-k", duration.asSlurm(args.timelimit_min)]
    if args.timelimit_max:
        opts += ["-K", duration.asSlurm(args.timelimit_max)]
    if args.partition:
        opts += ["-r", ",".join(args.partition)]
    if args.nodelist:
        opts += ["-N", ",".join(args.nodelist)]
    if args.clusters:
        opts += ["-M", ",".join(args.clusters)]
    if args.account:
        opts += ["-A", ",".join(args.account)]
    if args.qos:
        opts += ["-q", ",".join(args.qos)]
    if args.jobid:
        opts += ["-j", ",".join(map(str, args.jobid))]

    if args.name:
        vec = [["--name", i] for i in args.name]
        opts += [num for elem in vec for num in elem]

    if args.U:
        opts += ["-u", pwd.getpwuid(os.getuid())[0]]
    elif args.user and args.uid:
        opts += ["-u", ",".join(args.user) + "," + ",".join(args.uid)]
    elif args.user:
        opts += ["-u", ",".join(args.user)]
    elif args.uid:
        opts += ["-u", ",".join(args.uid)]

    if args.group and args.gid:
        opts += ["-u", ",".join(args.group) + "," + ",".join(args.gid)]
    elif args.group:
        opts += ["-u", ",".join(args.group)]
    elif args.gid:
        opts += ["-u", ",".join(args.gid)]

    lines = _read(" ".join(["sacct"] + opts))

    if extra:
        op = [i for i in opts] + ["--format", ",".join(extra)]
        op.remove("-l")
        ex = _read(" ".join(["sacct"] + op))
        for i in range(len(lines)):
            lines[i] = {**lines[i], **ex[i]}

    if args.root:
        lines = [i for i in lines if not os.path.relpath(i["WorkDir"], args.root).startswith("..")]
        for line in lines:
            line["WorkDir"] = os.path.relpath(line["WorkDir"], args.root)
    elif "WorkDir" in extra:
        if args.abspath:
            for line in lines:
                line["WorkDir"] = os.path.abspath(line["WorkDir"])
        elif args.relpath:
            for line in lines:
                line["WorkDir"] = os.path.relpath(line["WorkDir"])
        else:
            for line in lines:
                if len(os.path.relpath(line["WorkDir"]).split("../")) < 3:
                    line["WorkDir"] = os.path.relpath(line["WorkDir"])

    if args.infer:
        if not args.allocations:
            raise ValueError("Cannot infer extra data without --allocations.")
        opts.remove("-X")
        ex = _read(" ".join(["sacct"] + opts))
        data = defaultdict(dict)
        for line in ex:
            if "." not in line["JobID"]:
                continue
            jobid, key = line["JobID"].split(".")
            data[key][jobid] = line
        for line in lines:
            for key in line:
                if len(line[key]) == 0:
                    if line["JobID"] in data[args.infer]:
                        line[key] = data[args.infer][line["JobID"]][key]

    if args.state:
        alias = {
            "r": "RUNNING",
            "running": "RUNNING",
            "cd": "COMPLETED",
            "c": "COMPLETED",
            "completed": "COMPLETED",
            "f": "FAILED",
            "failed": "FAILED",
            "to": "TIMEOUT",
            "timeout": "TIMEOUT",
            "rs": "RESIZING",
            "resizing": "RESIZING",
            "dl": "DEADLINE",
            "deadline": "DEADLINE",
            "nf": "NODE_FAIL",
            "node_fail": "NODE_FAIL",
        }
        keys = []
        for i in args.state:
            for j in i.split(","):
                keys.append(j)
                if j not in alias:
                    alias[j] = j
        key = "State"
        fields = [alias[i] for i in keys]
        lines = [
            line
            for line in lines
            if sum(1 if re.match(n, str(line[key]), re.IGNORECASE) else 0 for n in fields)
        ]

    # shorten state
    shorten = {"COMPLETED": "C", "RUNNING": "R", "FAILED": "F"}
    shorten_state = True
    for i in list({i["State"] for i in lines}):
        if i not in shorten:
            shorten_state = False
            break
    if shorten_state:
        for line in lines:
            line["State"] = shorten[line["State"]]

    if args.sort:
        lookup = {i.upper(): i for i in lines[0].keys()}
        idx = np.lexsort([[line[lookup[key.upper()]] for line in lines] for key in args.sort])
        if args.reverse:
            idx = idx[::-1]
        lines = [lines[i] for i in idx]
    elif args.reverse:
        lines = [i for i in lines[::-1]]

    if args.json:
        for line in lines:
            line = {k: v for k, v in line.items() if len(v) > 0}
            json_object = json.dumps(line, indent=4)
            print(json_object)
        return

    default = [
        "JobID",
        "User",
        "Account",
        "JobName",
        "State",
        "Partition",
        "Elapsed",
        "ExitCode",
        "Reason",
        "Priority",
        "CPUTime",
        "AveCPU",
        "AveDiskRead",
        "AveDiskWrite",
        "MaxVMSizeNode",
        "MaxVMSize",
    ] + extra
    default = sorted(set(default), key=lambda x: default.index(x))

    columns = [{"key": key, "width": len(key), "align": ">", "priority": True} for key in default]
    header = {key: key for key in default}
    columns[default.index("AveCPU")]["priority"] = False
    columns[default.index("AveDiskRead")]["priority"] = False
    columns[default.index("AveDiskWrite")]["priority"] = False
    columns[default.index("MaxVMSize")]["priority"] = False
    columns[default.index("MaxVMSizeNode")]["priority"] = False
    columns[default.index("JobName")]["align"] = "<"
    columns[default.index("State")]["align"] = "<"
    columns[default.index("Partition")]["align"] = "<"
    if "WorkDir" in extra:
        columns[default.index("WorkDir")]["align"] = "<"

    for i in range(len(lines)):
        for key in ["Elapsed", "CPUTime", "AveCPU"]:
            if key in lines[i]:
                lines[i][key] = rich.Duration(lines[i][key])
        for key in ["AveDiskRead", "AveDiskWrite", "MaxVMSize", "MaxRSS"]:
            if key in lines[i]:
                lines[i][key] = rich.Memory(lines[i][key])

    if len(lines) > 0:
        keep = [False for _ in default]
        for i, key in enumerate(default):
            for line in lines:
                if key not in line:
                    continue
                if len(line[key]) > 0:
                    keep[i] = True
                    continue
        columns = [c for c, k in zip(columns, keep) if k]
        default = [c for c, k in zip(default, keep) if k]

    if shorten_state:
        header["State"] = "ST"
        columns[default.index("State")]["width"] = 2
        header["ExitCode"] = "exit"
        columns[default.index("ExitCode")]["width"] = 4

    table.print_columns(lines, columns, header, sep=args.sep, no_truncate=args.no_truncate)


def _Gacct_catch():
    Gacct(sys.argv[1:])



import re

from . import rich


def colors(theme=None):
    r"""
    Return dictionary of colors.

    .. code-block:: python

        {
            'selection' : '...',
            'free'      : '...',
            'error'     : '...',
            'warning'   : '...',
            'low'       : '...',
        }

    :options:

        **theme** ([``'dark'``] | ``<str>``)
            Select color-theme.
    """

    if theme == "dark":
        return {
            "selection": "1;32;40",
            "free": "1;32",
            "error": "9;31",
            "warning": "1;31",
            "low": "1;36",
        }

    return {
        "selection": "",
        "free": "",
        "error": "",
        "warning": "",
        "low": "",
    }


def read(data=None):
    r"""
    Read ``sinfo -o "%all"``.

    :options:

        **data** (``<str>``)
            For debugging: specify the output of ``sinfo -o "%all"`` as string.

    :returns:

        **lines** ``<list<dict>>``
            A list of dictionaries, that contain the different fields. All data are strings.
    """

    import subprocess

    # get live info
    if data is None:
        data = subprocess.check_output('sinfo -o "%all"', shell=True).decode("utf-8")

    # extract the header and the info
    head, data = data.split("\n", 1)
    data = list(filter(None, data.split("\n")))

    # get the field-names
    head = head.split("|")

    # convert to list of dictionaries
    # - initialize
    lines = []
    # - loop over lines
    for line in data:
        # -- initialize empty dictionary
        info = {}
        # -- fill dictionary
        for key, val in zip(head, line.split("|")):
            if len(key.strip()) > 0:
                info[key.strip()] = val.strip()
        # -- store to list of lines
        lines += [info]

    # return output
    return lines


def cpu_score(CPU_LOAD, CPUS_A, **kwargs):

    if not CPU_LOAD.isnumeric():
        return rich.Float("")
    if int(CPUS_A) <= 0:
        return rich.Float("")

    return rich.Float(float(CPU_LOAD) / float(CPUS_A), precision=2)


def mem_score(MEMORY, FREE_MEM, CPUS_A, CPUS_T, **kwargs):

    if not FREE_MEM.isnumeric():
        return rich.Float("")
    if not CPUS_T.isnumeric():
        return rich.Float("")
    if int(MEMORY) <= 0:
        return rich.Float("")
    if int(CPUS_A) <= 0:
        return rich.Float("")

    MEM_USED = float(MEMORY) - float(FREE_MEM)

    return rich.Float(float(MEM_USED) / float(MEMORY) * float(CPUS_T) / float(CPUS_A), precision=2)


def interpret(lines, theme=colors()):
    r"""
    Interpret the output of ``GooseSLURM.sinfo.read``. All fields are converted to the
    ``GooseSLURM.rich`` classes adding useful colors in the process.

    :arguments:

        **lines** ``<list<dict>>``
            The output of ``GooseSLURM.sinfo.read``

    :options:

        **theme** (``<dict>``)
            The color-theme, as selected by ``GooseSLURM.sinfo.colors``.

    :returns:

        **lines** (``<list<dict>>``)
            A list of dictionaries, that contain the different fields. All data are
            ``GooseSLURM.rich.String`` or derived types.
    """

    # interpret input as list
    if not isinstance(lines, list):
        lines = [lines]

    # loop over all lines
    for line in lines:

        # convert fields to string
        for key in line:
            if not isinstance(line[key], rich.String):
                line[key] = rich.String(line[key])

                # covert to float
        for key in ["CPU_LOAD"]:
            line[key] = rich.Float(str(line[key]))

        # "days-hours:mins:secs" (e.g. "1-4:18:13") -> seconds
        for key in ["TIMELIMIT"]:
            line[key] = rich.Duration(str(line[key]))

        # convert memory (e.g. "4G") -> bytes
        for key in ["MEMORY", "FREE_MEM"]:
            line[key] = rich.Memory(str(line[key]), default_unit=1e6)

        # CPUs
        # - split: allocated/idle/other/total
        a, i, o, t = str(line["CPUS(A/I/O/T)"]).split("/")
        # - store extra fields
        line["CPUS_A"] = rich.Integer(a)
        line["CPUS_I"] = rich.Integer(i)
        line["CPUS_O"] = rich.Integer(o)
        line["CPUS_T"] = rich.Integer(t)

        # compute scores
        line["CPU_RELJOB"] = cpu_score(**line)
        line["MEM_RELJOB"] = mem_score(**line)

        # default: CPUs down/online
        line["CPUS_D"] = rich.Integer(0)
        line["CPUS_O"] = rich.Integer(str(line["CPUS_T"]))

        # node down: mark all fields
        if (
            re.match(r"down,*", str(line["STATE"]))
            or re.match(r"maint.*", str(line["STATE"]))
            or re.match(r"drain.*", str(line["STATE"]))
        ):

            line["CPUS_I"] = rich.Integer(0)
            line["CPUS_D"] = rich.Integer(str(line["CPUS_T"]))
            line["CPUS_O"] = rich.Integer(0)

            for key in line:
                line[key].color = theme["error"]

        # highlight 'scores'
        if int(line["CPUS_I"]) > 0:
            line["CPUS_I"].color = theme["free"]
        if float(line["CPU_RELJOB"]) > 1.05:
            line["CPU_RELJOB"].color = theme["warning"]
        elif float(line["CPU_RELJOB"]) < 0.95:
            line["CPU_RELJOB"].color = theme["low"]

        # highlight 'scores'
        if float(line["MEM_RELJOB"]) > 0.9:
            line["MEM_RELJOB"].color = theme["warning"]

    return lines


def read_interpret(data=None, theme=colors()):
    r"""
    Read and interpret ``sinfo -o "%all"``.

    :returns:

        **lines** (``<list<dict>>``)
            A list of dictionaries, that contain the different fields. All data are
            ``GooseSLURM.rich.String`` or derived types.
    """

    return interpret(read(data), theme)



from . import duration
from . import memory


class String:
    r"""
    Rich string.

    .. note::

        All options are attributes, that can be modified at all times.

    :options:

        **data** (``<str>`` | ``None``)
            The data.

        **width** ([``None``] | ``<int>``)
            Print width (formatted print only).

        **color** ([``None``] | ``<str>``)
            Print color, e.g. "1;32" for bold green (formatted print only).

        **align** ([``'<'``] | ``'>'``)
            Print alignment (formatted print only).

        **dummy** ([``0``] | ``<int>`` | ``<float>``)
            Dummy numerical value.

    :methods:

        **A.format()**
            Formatted string.

        **str(A)**
            Unformatted string.

        **A.isnumeric()**
            Return if the "data" is numeric.

        **int(A)**
            Dummy integer.

        **float(A)**
            Dummy float.
    """

    def __init__(self, data, width=None, align="<", color=None, dummy=0):

        self.data = data
        self.width = width
        self.color = color
        self.align = align
        self.dummy = dummy

    def format(self):
        r"""
        Return formatted string: align/width/color are applied.
        """

        if self.width and self.color:
            fmt = "\x1b[{color:s}m{{0:{align:s}{width:d}.{width:d}s}}\x1b[0m".format(
                **self.__dict__
            )
        elif self.width:
            fmt = "{{0:{align:s}{width:d}.{width:d}s}}".format(**self.__dict__)
        elif self.color:
            fmt = "\x1b[{color:s}m{{0:{align:s}s}}\x1b[0m".format(**self.__dict__)
        else:
            fmt = "{{0:{align:s}s}}".format(**self.__dict__)

        return fmt.format(str(self))

    def isnumeric(self):
        r"""
        Return if the "data" is numeric : always zero for this class.
        """

        return False

    def __str__(self):

        return str(self.data)

    def __len__(self):

        return len(str(self))

    def __int__(self):

        return int(self.dummy)

    def __float__(self):

        return float(self.dummy)

    def __repr__(self):

        return str(self)

    def __lt__(self, other):

        return str(self) < str(other)


class Integer(String):
    r"""
    Rich integer.

    .. note::

        All options are attributes, that can be modified at all times.

    :options:

        **data** (``<str>`` | ``None``)
            The data.

        **width** ([``None``] | ``<int>``)
            Print width (formatted print only).

        **color** ([``None``] | ``<str>``)
            Print color, e.g. "1;32" for bold green (formatted print only).

        **align** ([``'<'``] | ``'>'``)
            Print alignment (formatted print only).

        **dummy** ([``0``] | ``<int>`` | ``<float>``)
            Dummy numerical value, used in case of non-numerical ``data``.

    :methods:

        **A.format()**
            Formatted string.

        **str(A)**
            Unformatted string.

        **A.isnumeric()**
            Return if the "data" is numeric.

        **int(A)**
            Return ``data`` as integer (``dummy`` is returned if ``data`` is not numeric).

        **float(A)**
            Return ``data`` as float (``dummy`` is returned if ``data`` is not numeric).
    """

    def __init__(self, data, **kwargs):

        try:
            data = int(data)
        except BaseException:
            pass

        super().__init__(data, **kwargs)

    def isnumeric(self):
        r"""
        Return if the "data" is numeric : always zero for this class.
        """

        if isinstance(self.data, int):
            return True
        if isinstance(self.data, float):
            return True

        return False

    def __int__(self):

        if isinstance(self.data, int):
            return self.data
        if isinstance(self.data, float):
            return int(self.data)

        return int(self.dummy)

    def __float__(self):

        if isinstance(self.data, int):
            return float(self.data)
        if isinstance(self.data, float):
            return self.data

        return float(self.dummy)

    def __lt__(self, other):

        if self.isnumeric() and other.isnumeric():
            return int(self) < int(other)
        elif self.isnumeric():
            return 0
        elif other.isnumeric():
            return -1

        return self.data < other.data


class Float(String):
    r"""
    Rich float.

    .. note::

        All options are attributes, that can be modified at all times.

    :options:

        **data** (``<str>`` | ``None``)
            The data.

        **width** ([``None``] | ``<int>``)
            Print width (formatted print only).

        **color** ([``None``] | ``<str>``)
            Print color, e.g. "1;32" for bold green (formatted print only).

        **align** ([``'<'``] | ``'>'``)
            Print alignment (formatted print only).

        **precision** ([``2``] | ``<int>``)
            Print precision (formatted print only).

        **dummy** ([``0``] | ``<int>`` | ``<float>``)
            Dummy numerical value, used in case of non-numerical ``data``.

    :methods:

        **A.format()**
            Formatted string.

        **str(A)**
            Unformatted string.

        **A.isnumeric()**
            Return if the "data" is numeric.

        **int(A)**
            Return ``data`` as integer (``dummy`` is returned if ``data`` is not numeric).

        **float(A)**
            Return ``data`` as float (``dummy`` is returned if ``data`` is not numeric).
    """

    def __init__(self, data, **kwargs):

        try:
            data = float(data)
        except BaseException:
            pass

        self.precision = kwargs.pop("precision", 2)

        super().__init__(data, **kwargs)

    def isnumeric(self):
        r"""
        Return if the "data" is numeric : always zero for this class.
        """

        if isinstance(self.data, float):
            return True
        if isinstance(self.data, int):
            return True

        return False

    def __str__(self):

        if not self.isnumeric():
            return self.data

        return "{{0:.{precision:d}f}}".format(**self.__dict__).format(self.data)

    def __int__(self):

        if isinstance(self.data, float):
            return int(self.data)
        if isinstance(self.data, int):
            return self.data

        return int(self.dummy)

    def __float__(self):

        if isinstance(self.data, float):
            return self.data
        if isinstance(self.data, int):
            return float(self.data)

        return float(self.dummy)

    def __lt__(self, other):

        if self.isnumeric() and other.isnumeric():
            return float(self) < float(other)
        elif self.isnumeric():
            return 0
        elif other.isnumeric():
            return -1

        return self.data < other.data


class Duration(Integer):
    r"""
    Rich duration (seconds).

    .. note::

        All options are attributes, that can be modified at all times.

    :options:

        **data** (``<str>`` | ``None``)
            The data.

        **width** ([``None``] | ``<int>``)
            Print width (formatted print only).

        **color** ([``None``] | ``<str>``)
            Print color, e.g. "1;32" for bold green (formatted print only).

        **align** ([``'<'``] | ``'>'``)
            Print alignment (formatted print only).

        **precision** ([``None``] | ``<int>``)
            Print precision to use for the conversion (formatted print only).
            ``None`` means automatic precision.
            See ``GooseSLURM.duration.asHuman``.

        **dummy** ([``0``] | ``<int>`` | ``<float>``)
            Dummy numerical value, used in case of non-numerical ``data``.

    :methods:

        **A.format()**
            Formatted string, after unit conversion.

        **str(A)**
            Unformatted string, after unit conversion.

        **A.isnumeric()**
            Return if the "data" is numeric.

        **int(A)**
            Return ``data`` as integer (``dummy`` is returned if ``data`` is not numeric).

        **float(A)**
            Return ``data`` as float (``dummy`` is returned if ``data`` is not numeric).
    """

    def __init__(self, data, **kwargs):

        data = duration.asSeconds(data, default=data)

        self.precision = kwargs.pop("precision", None)

        super().__init__(data, **kwargs)

    def __str__(self):

        if not self.isnumeric():
            return self.data

        return duration.asHuman(self.data, self.precision)


class Memory(Integer):
    r"""
    Rich memory (bytes).

    .. note::

        All options are attributes, that can be modified at all times.

    :options:

        **data** (``<str>`` | ``None``)
            The data.

        **width** ([``None``] | ``<int>``)
            Print width (formatted print only).

        **color** ([``None``] | ``<str>``)
            Print color, e.g. "1;32" for bold green (formatted print only).

        **align** ([``'<'``] | ``'>'``)
            Print alignment (formatted print only).

        **precision** ([``None``] | ``<int>``)
            Print precision to use for the conversion (formatted print only).
            ``None`` means automatic precision.
            See ``GooseSLURM.memory.asHuman``.

        **dummy** ([``0``] | ``<int>`` | ``<float>``)
            Dummy numerical value, used in case of non-numerical ``data``.

        **default_unit** (``int``)
            The unit to assume if no unit if specified (specify the number of bytes).

    :methods:

        **A.format()**
            Formatted string, after unit conversion.

        **str(A)**
            Unformatted string, after unit conversion.

        **A.isnumeric()**
            Return if the "data" is numeric.

        **int(A)**
            Return ``data`` as integer (``dummy`` is returned if ``data`` is not numeric).

        **float(A)**
            Return ``data`` as float (``dummy`` is returned if ``data`` is not numeric).
    """

    def __init__(self, data, **kwargs):

        if "default_unit" in kwargs:
            data = memory.asBytes(data, default=data, default_unit=kwargs.pop("default_unit"))
        else:
            data = memory.asBytes(data, default=data)

        self.precision = kwargs.pop("precision", None)

        super().__init__(data, **kwargs)

    def __str__(self):

        if not self.isnumeric():
            return self.data

        return memory.asHuman(self.data, self.precision)



from . import rich


def colors(theme=None):
    r"""
    Return dictionary of colors.

    .. code-block:: python

      {
        'selection' : '...',
        'queued'    : '...',
      }

    :options:

      **theme** ([``'dark'``] | ``<str>``)
        Select color-theme.
    """

    if theme == "dark":
        return {
            "selection": "1;32;40",
            "queued": "0;37",
        }

    return {
        "selection": "",
        "queued": "",
    }


def read(data=None):
    r"""
    Read ``squeue -o "%all"``.

    :options:

      **data** (``<str>``)
        For debugging: specify the output of ``squeue -o "%all"`` as string.

    :returns:

      **lines** ``<list<dict>>``
        A list of dictionaries, that contain the different fields. All data are strings.
    """

    import subprocess

    # get live info
    if data is None:
        data = subprocess.check_output('squeue -o "%all"', shell=True).decode("utf-8")

    # extract the header and the info
    head, data = data.split("\n", 1)
    data = list(filter(None, data.split("\n")))

    # get the field-names
    head = head.split("|")

    # convert name (bug in slurm)
    # Bug #4948, https://bugs.schedmd.com
    idx = [i for i, key in enumerate(head) if key.strip() == "USER"]
    if len(idx) > 1:
        head[idx[1]] = "USER_ID"

    # convert to list of dictionaries
    # - initialize
    lines = []
    # - loop over lines
    for line in data:
        # -- initialize empty dictionary
        info = {}
        # -- fill dictionary
        for key, val in zip(head, line.split("|")):
            if len(key.strip()) > 0:
                info[key.strip()] = val.strip()
        # -- store to list of lines
        lines += [info]

    # return output
    return lines


def interpret(lines, now=None, theme=colors()):
    r"""
    Interpret the output of ``GooseSLURM.squeue.read``. All fields are converted to the
    ``GooseSLURM.rich`` classes adding useful colors in the process.

    :arguments:

      **lines** ``<list<dict>>``
        The output of ``GooseSLURM.squeue.read``

    :options:

      **theme** (``<dict>``)
        The color-theme, as selected by ``GooseSLURM.squeue.colors``.

    :returns:

      **lines** (``<list<dict>>``)
        A list of dictionaries, that contain the different fields. All data are
        ``GooseSLURM.rich.String`` or derived types.
    """

    import time

    # read time
    if now is None:
        now = time.mktime(time.localtime())

    # interpret input as list
    if not isinstance(lines, list):
        lines = [lines]

    # loop over all lines
    for line in lines:

        # convert to integer
        for key in ["CPUS", "NODES"]:
            line[key] = rich.Integer(line[key])

        # "year-month-dayThour:minute:second" (e.g. "2017-11-05T19:09:53") -> seconds from now
        for key in ["START_TIME", "SUBMIT_TIME"]:
            try:
                line[key] = rich.Duration(
                    int(now - time.mktime(time.strptime(line[key], "%Y-%m-%dT%H:%M:%S")))
                )
            except BaseException:
                line[key] = rich.Duration(line[key])

        # "days-hours:mins:secs" (e.g. "1-4:18:13") -> seconds
        for key in ["TIME_LIMIT", "TIME_LEFT", "TIME"]:
            line[key] = rich.Duration(line[key])

        # convert memory (e.g. "4G") -> bytes
        for key in ["MIN_MEMORY"]:
            line[key] = rich.Memory(line[key])

        # specialize number of CPUS
        if str(line["ST"]) == "R":
            line["CPUS_R"] = rich.Integer(int(line["CPUS"]))
            line["CPUS_PD"] = rich.Integer(0)
        else:
            line["CPUS_R"] = rich.Integer(0)
            line["CPUS_PD"] = rich.Integer(int(line["CPUS"]))

        # convert remaining fields to string
        for key in line:
            if not isinstance(line[key], rich.String):
                line[key] = rich.String(line[key])

        # highlight queued jobs
        if str(line["ST"]) == "PD":
            for key in line:
                line[key].color = theme["queued"]

    return lines


def read_interpret(data=None, now=None, theme=colors()):
    r"""
    Read and interpret ``squeue -o "%all"``.

    :returns:

      **lines** (``<list<dict>>``)
        A list of dictionaries, that contain the different fields. All data are
        ``GooseSLURM.rich.String`` or derived types.
    """

    return interpret(read(data), now, theme)



from setuptools import find_packages
from setuptools import setup

setup(
    name="dummyslurm",
    license="MIT",
    author="Tom de Geus",
    author_email="tom@geus.me",
    description="Dummy slurm installation",
    long_description="Dummy slurm installation",
    keywords="SLURM",
    url="https://github.com/tdegeus/GooseSLURM",
    packages=find_packages(),
    install_requires=["PyYAML"],
    entry_points={
        "console_scripts": [
            "sacct = dummyslurm:sacct",
            "sbatch = dummyslurm:sbatch",
            "scancel = dummyslurm:scancel",
            "scontrol = dummyslurm:scontrol",
            "squeue = dummyslurm:squeue",
        ]
    },
)



from . import rich


def colors(theme=None):
    r"""
    Return dictionary of colors.

    .. code-block:: python

        {
          'selection' : '...',
        }

    :options:

        **theme** ([``'dark'``] | ``<str>``)
            Select color-theme.
    """

    if theme == "dark":
        return {
            "selection": "1;32;40",
        }

    return {
        "selection": "",
    }


def read(data=None):
    r"""
    Read ``ps -eo pid,user,rss,%cpu,command``.

    :options:

        **data** (``<str>``)
            For debugging: specify the output of ``ps -eo pid,user,rss,%cpu,command`` as string.

    :returns:

        **lines** ``<list<dict>>``
            A list of dictionaries, that contain the different fields. All data are strings.
    """

    import subprocess

    # get live info
    if data is None:
        data = subprocess.check_output("ps -eo pid,user,rss,%cpu,command", shell=True).decode(
            "utf-8"
        )

    # extract the header and the info
    header, data = data.split("\n", 1)
    data = list(filter(None, data.split("\n")))
    header = list(filter(None, header.split(" ")))

    # convert to list of dictionaries
    # - initialize
    lines = []
    # - loop over lines
    for line in data:
        # -- decode line with information
        d = ["" for i in header]
        line = line.strip()
        d[0], line = line.split(" ", 1)
        line = line.strip()
        d[1], line = line.split(" ", 1)
        line = line.strip()
        d[2], line = line.split(" ", 1)
        line = line.strip()
        d[3], line = line.split(" ", 1)
        d[4] = line.strip()
        # -- initialize empty dictionary
        info = {}
        # -- fill dictionary
        for key, val in zip(header, d):
            info[key] = val
        # -- store to list of lines
        lines += [info]

    # return output
    return lines


def interpret(lines, theme=colors()):
    r"""
    Interpret the output of ``GooseSLURM.ps.read``. All fields are converted to the
    ``GooseSLURM.rich`` classes adding useful colors in the process.

    :arguments:

        **lines** ``<list<dict>>``
            The output of ``GooseSLURM.ps.read``

    :options:

        **theme** (``<dict>``)
            The color-theme, as selected by ``GooseSLURM.ps.colors``.

    :returns:

        **lines** (``<list<dict>>``)
            A list of dictionaries, that contain the different fields. All data are
            ``GooseSLURM.rich.String`` or derived types.
    """

    # interpret input as list
    if not isinstance(lines, list):
        lines = [lines]

    # loop over all lines
    for line in lines:

        # custom conversion
        for key in ["%CPU"]:
            line[key] = rich.Float(line[key], precision=2)

        # custom conversion
        for key in ["RSS"]:
            line[key] = rich.Memory(line[key], default_unit=1e3)

        # convert remaining fields to string
        for key in line:
            if not isinstance(line[key], rich.String):
                line[key] = rich.String(line[key])

    return lines


def read_interpret(data=None, theme=colors()):
    r"""
    Read and interpret ``ps -eo pid,user,rss,%cpu,command``.

    :returns:

        **lines** (``<list<dict>>``)
            A list of dictionaries, that contain the different fields. All data are
            ``GooseSLURM.rich.String`` or derived types.
    """

    return interpret(read(data), theme)



"""Ginfo
    Summarize the status of the compute nodes (wrapper around "sinfo").

    The following scores are computed of each node:

    *   CPU% : The CPU load of the node relative to the number of jobs (``cpu_load / cpu_used``).
        Should always be ~1, anything else usually signals misuse.

    *   Mem% : the amount of used memory relative to the average memory available per job
        (``(mem_used / cpu_used) / (mem_tot / cpu_tot)``).
        Can be > 1 for (several) heavy memory consumption jobs,
        but in principle any value is possible.

Usage:
    Ginfo [options]

Options:
    -U
        Limit output to the current user.

    -u, --user=<NAME>
        Limit output to user(s).
        Option may be repeated. Search by regex.

    -j, --jobid=<NAME>
        Limit output to job-id(s).
        Option may be repeated. Search by regex.

    --host=<NAME>
        Limit output to host(s).
        Option may be repeated. Search by regex.

    -f, --cfree=<NAME>
        Limit output to free CPU(s).
        Option may be repeated. Search by regex.

    -p, --partition=<NAME>
        Limit output to partition(s).
        Option may be repeated. Search by regex.

    -s, --sort=<NAME>
        Sort by field.
        Option may be repeated. Use header names.

    -r, --reverse
        Reverse sort.

    -o, --output=<NAME>
        Select output columns.
        Option may be repeated. Use header names.

    -S, --summary
        Print only summary.

    --no-header
        Suppress header.

    --no-truncate
        Print full columns, do not truncate based on terminal width.

    --width=<N>
        Set line-width (otherwise taken as terminal width).

    --colors=<NAME>
        Select color scheme from: "none", "dark". [default: "dark"]

    -l, --list
        Print selected column as list.

    --sep=<NAME>
        Set column separator. [default: " "]

    --long
        Print full information (each column is printed as a line).

    --debug=<FILE> <FILE>
        Debug: read ``sinfo -o "%all"`` and  ``squeue -o "%all"`` from file.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseSLURM
"""
import argparse
import os
import pwd
import re
import sys

import numpy as np

from .. import rich
from .. import sinfo
from .. import squeue
from .. import table
from .. import version


def main():

    # -- parse command line arguments --

    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-U", action="store_true")
    parser.add_argument("-u", "--user", type=str, action="append", default=[])
    parser.add_argument("-j", "--jobid", type=str, action="append")
    parser.add_argument("--host", type=str, action="append")
    parser.add_argument("-c", "--cfree", type=str, action="append")
    parser.add_argument("-p", "--partition", type=str, action="append")
    parser.add_argument("-s", "--sort", type=str, action="append")
    parser.add_argument("-r", "--reverse", action="store_true")
    parser.add_argument("-o", "--output", type=str, action="append")
    parser.add_argument("-S", "--summary", action="store_true")
    parser.add_argument("--no-header", action="store_true")
    parser.add_argument("--no-truncate", action="store_true")
    parser.add_argument("--width", type=int)
    parser.add_argument("--colors", type=str, default="dark")
    parser.add_argument("-l", "--list", action="store_true")
    parser.add_argument("--sep", type=str, default=" ")
    parser.add_argument("--long", action="store_true")
    parser.add_argument("--debug", type=str, nargs=2)
    parser.add_argument("--version", action="version", version=version)
    args = vars(parser.parse_args())

    # -------------------------------- field-names and print settings --------

    # conversion map: default field-names -> custom field-names
    alias = {
        "HOSTNAMES": "Host",
        "CPUS_T": "CPUs",
        "CPUS_I": "Cfree",
        "CPUS_D": "Cdown",
        "CPUS_O": "Con",
        "CPU_RELJOB": "CPU%",
        "PARTITION": "Partition",
        "MEMORY": "Mem",
        "FREE_MEM": "Mfree",
        "MEM_RELJOB": "Mem%",
        "TIMELIMIT": "Tlim",
        "STATE": "State",
    }

    # conversion map: custom field-names -> default field-names
    aliasInv = {alias[key].upper(): key for key in alias}

    # rename command line options -> default field-names
    for key in [key for key in args]:
        if key.upper() in aliasInv:
            args[aliasInv[key.upper()]] = args.pop(key)

    # print settings of all columns
    # - "width"   : minimum width, adapted to print width (min_width <= width <= real_width)
    # - "align"   : alignment of the columns (except the header)
    # - "priority": priority of column expansing, columns marked "True" are expanded first
    columns = [
        {"key": "HOSTNAMES", "width": 4, "align": "<", "priority": True},
        {"key": "CPUS_T", "width": 4, "align": ">", "priority": True},
        {"key": "CPUS_I", "width": 5, "align": ">", "priority": True},
        {"key": "CPU_RELJOB", "width": 4, "align": ">", "priority": True},
        {"key": "MEMORY", "width": 3, "align": ">", "priority": True},
        {"key": "FREE_MEM", "width": 5, "align": ">", "priority": True},
        {"key": "MEM_RELJOB", "width": 4, "align": ">", "priority": True},
        {"key": "PARTITION", "width": 9, "align": "<", "priority": True},
        {"key": "TIMELIMIT", "width": 4, "align": ">", "priority": False},
        {"key": "STATE", "width": 5, "align": "<", "priority": False},
    ]

    # header
    header = {
        column["key"]: rich.String(alias[column["key"]], align=column["align"])
        for column in columns
    }

    # print settings for the summary
    columns_summary = [
        {"key": "PARTITION", "width": 9, "align": "<", "priority": True},
        {"key": "CPUS_T", "width": 4, "align": ">", "priority": True},
        {"key": "CPUS_O", "width": 5, "align": ">", "priority": True},
        {"key": "CPUS_D", "width": 5, "align": ">", "priority": True},
        {"key": "CPUS_I", "width": 5, "align": ">", "priority": True},
        {"key": "CPU_RELJOB", "width": 4, "align": ">", "priority": True},
        {"key": "MEM_RELJOB", "width": 4, "align": ">", "priority": True},
    ]

    # header
    header_summary = {
        column["key"]: rich.String(alias[column["key"]], align=column["align"])
        for column in columns_summary
    }

    # select color theme
    theme = sinfo.colors(args["colors"].lower())

    # -- load the output of "sinfo" --

    if not args["debug"]:

        lines = sinfo.read_interpret(theme=theme)

    else:

        lines = sinfo.read_interpret(
            data=open(args["debug"][0]).read(),
            theme=theme,
        )

    # ----------------------------- limit based on command-line options ------

    for key in ["HOSTNAMES", "PARTITION", "CPUS_I"]:

        if args[key]:

            # limit data
            lines = [
                line
                for line in lines
                if sum(1 if re.match(n, str(line[key])) else 0 for n in args[key])
            ]

            # color-highlight selected columns
            # - apply to all remaining lines
            for line in lines:
                line[key].color = theme["selection"]
            # - apply to the header
            header[key].color = theme["selection"]

    # -- support function used below --

    # if needed, convert 'name[10-14,16]' to 'list(name10, name11, name12, name13, name14, name16)'
    def expand_nodelist(text):

        # try to split 'name', '[10-14]'
        match = list(filter(None, re.split(r"(\[[^\]]*\])", text)))

        # no split made: no need to interpret anything, return as list
        if len(match) == 1:
            return [text]

        # split in variables
        name, numbers = match

        # remove brackets '[10-14,16]' -> '10-14,16'
        numbers = numbers[1:-1]

        # split '10-14,16' -> list('10-14', '16')
        numbers = numbers.split(",")

        # allocate output
        nodes = []

        # expand if needed
        for number in numbers:

            # '16' -> 'name16'
            if len(number.split("-")) == 1:

                # copy to list
                nodes += [name + number]

            # '10-14' -> list('name10', 'name11', 'name12', 'name13', 'name14')
            else:

                # get start and end numbers
                start, end = number.split("-")

                # expand between beginning and end
                nodes += [
                    name + ("%0" + str(len(start)) + "d") % i
                    for i in range(int(start), int(end) + 1)
                ]

        # return output
        return nodes

    # -- limit to users --

    # handle 'alias' options
    if args["U"]:
        args["user"] += [pwd.getpwuid(os.getuid())[0]]

    # apply filter
    if args["user"] or args["jobid"]:

        # get list of jobs
        # ----------------

        # read
        if not args["debug"]:

            jobs = squeue.read_interpret()

        else:

            jobs = squeue.read_interpret(
                data=open(args["debug"][1]).read(),
                now=os.path.getctime(args["debug"][1]),
            )

        # limit to running jobs
        jobs = [j for j in jobs if str(j["ST"]) == "R"]

        # limit to users' jobs
        if args["user"]:
            jobs = [
                str(j["NODELIST"])
                for j in jobs
                if sum(1 if re.match(n, str(j["USER"])) else 0 for n in args["user"])
            ]

        # limit to specific jobs
        if args["jobid"]:
            jobs = [
                str(j["NODELIST"])
                for j in jobs
                if sum(1 if re.match(n, str(j["JOBID"])) else 0 for n in args["jobid"])
            ]

        # get list of nodes for the users' jobs
        # --

        # allocate list of nodes
        nodes = []

        # loop over jobs
        for job in jobs:

            # simple name (e.g. 'f123') -> add to list
            if len(job.split(",")) == 1:
                nodes += expand_nodelist(job)
                continue

            # split all array jobs, e.g.
            # g117,g[123-456],f[023-025] -> ('g117,g', '[123-456]', ',f', '[023-025]')
            match = list(filter(None, re.split(r"(\[[^\]]*\])", job)))

            # loop over arrays
            for name, numbers in zip(match[0::2], match[1::2]):

                # strip plain jobs that are still prepending the array
                name = name.split(",")
                # add plain jobs to node-list
                nodes += name[:-1]
                # interpret all batch jobs and add to node-list
                nodes += expand_nodelist(name[-1] + numbers)

        # filter empty items
        nodes = list(filter(None, nodes))

        # limit data
        lines = [line for line in lines if str(line["HOSTNAMES"]) in nodes]

        # color-highlight selected columns
        # - apply to all remaining lines
        for line in lines:
            line["HOSTNAMES"].color = theme["selection"]
        # - apply to the header
        header["HOSTNAMES"].color = theme["selection"]

    # -- sort --

    if args["sort"]:
        sortkeys = [aliasInv[key.upper()] for key in args["sort"]]
    else:
        sortkeys = ["HOSTNAMES", "PARTITION"]

    idx = np.lexsort([[i[key] for i in lines] for key in sortkeys])
    if args["reverse"]:
        idx = idx[::-1]
    lines = [lines[i] for i in idx]

    # -- select columns --

    if args["output"]:

        keys = [aliasInv[key.upper()] for key in args["output"]]

        columns = [column for column in columns if column["key"] in keys]

    # -- print --

    if not args["summary"]:

        # optional: print all fields and quit
        if args["long"]:

            table.print_long(lines)

            sys.exit(0)

        # optional: print as list and quit
        elif args["list"]:

            # - only one field can be selected
            if len(columns) > 1:
                print("Only one field can be selected")
                sys.exit(1)

            # - print and quit
            table.print_list(lines, columns[0]["key"], args["sep"])

            sys.exit(0)

        # default: print columns
        else:

            table.print_columns(
                lines,
                columns,
                header,
                args["no_truncate"],
                args["sep"],
                args["width"],
                not args["no_header"],
            )

            sys.exit(0)

    # -- summarize information --

    # get names of the different partitions
    partitions = sorted({str(line["PARTITION"]) for line in lines})

    # start a new list of "node information", summed on the relevant nodes
    partitions = [{"PARTITION": rich.String(key)} for key in partitions]

    # loop over partitions
    for partition in partitions:

        # - isolate nodes for this partition
        N = [line for line in lines if str(line["PARTITION"]) == str(partition["PARTITION"])]

        # - get the CPU count
        partition["CPUS_T"] = rich.Integer(sum(int(line["CPUS_T"]) for line in N))
        partition["CPUS_O"] = rich.Integer(sum(int(line["CPUS_O"]) for line in N))
        partition["CPUS_D"] = rich.Integer(sum(int(line["CPUS_D"]) for line in N))
        partition["CPUS_I"] = rich.Integer(sum(int(line["CPUS_I"]) for line in N))

        # - initialize scores
        partition["CPU_RELJOB"] = rich.Float("")
        partition["MEM_RELJOB"] = rich.Float("")

        # - average load
        if len([1 for line in N if line["CPU_RELJOB"].isnumeric()]) > 0:
            partition["CPU_RELJOB"] = rich.Float(
                sum(float(line["CPU_RELJOB"]) for line in N if line["CPU_RELJOB"].isnumeric())
                / sum(1.0 for line in N if line["CPU_RELJOB"].isnumeric())
            )

        # - average memory consumption
        if len([1 for line in N if line["MEM_RELJOB"].isnumeric()]) > 0:
            partition["MEM_RELJOB"] = rich.Float(
                sum(float(line["MEM_RELJOB"]) for line in N if line["MEM_RELJOB"].isnumeric())
                / sum(1.0 for line in N if line["MEM_RELJOB"].isnumeric())
            )

        # - highlight 'scores'
        if int(partition["CPUS_I"]) > 0:
            partition["CPUS_I"].color = theme["free"]
        if float(partition["CPU_RELJOB"]) > 1.05:
            partition["CPU_RELJOB"].color = theme["warning"]
        elif float(partition["CPU_RELJOB"]) < 0.95:
            partition["CPU_RELJOB"].color = theme["low"]

    # rename field
    lines = partitions

    # -- sort --

    # default sort
    lines.sort(key=lambda line: line["PARTITION"], reverse=args["reverse"])

    # optional: sort by key(s)
    if args["sort"]:
        keys = [alias[column["key"]].upper() for column in columns_summary]
        args["sort"] = [key for key in args["sort"] if key.upper() in keys]

        idx = np.lexsort([[i[aliasInv[k.upper()]] for i in lines] for k in args["sort"]])
        if args["reverse"]:
            idx = idx[::-1]
        lines = [lines[i] for i in idx]

    # -- print --

    table.print_columns(
        lines,
        columns_summary,
        header_summary,
        args["no_truncate"],
        args["sep"],
        args["width"],
        not args["no_header"],
    )



"""Gps
    List memory usage per process.

        +--------------+------------------------------------------------+
        | Header       | Description                                    |
        +==============+================================================+
        | "PID"        | Process-id                                     |
        +--------------+------------------------------------------------+
        | "USER"       | Username                                       |
        +--------------+------------------------------------------------+
        | "MEM"        | Memory used                                    |
        +--------------+------------------------------------------------+
        | "%CPU"       | Fraction of CPU capacity used                  |
        +--------------+------------------------------------------------+
        | "COMMAND"    | Command                                        |
        +--------------+------------------------------------------------+

Usage:
    Gps [options]

Options:
    -U
        Limit processes to the current user.

    -u, --user=<NAME>
        Limit processes to user(s).
        Option may be repeated. Search by regex.

    -p, --pid=<NAME>
        Limit processes to process-id.
        Option may be repeated. Search by regex.

    -c, --command=<NAME>
        Limit processes to command.
        Option may be repeated. Search by regex.

    -s, --sort=<NAME>
        Sort by field (selected by the header name).

    -r, --reverse
        Reverse sort.

    -o, --output=<NAME>
        Select output columns.
        Option may be repeated. See description for header names.

    --no-header
        Suppress header.

    --no-truncate
        Print full columns, do not truncate based on terminal width.

    --width=<N>
        Set line-width (otherwise taken as terminal width).

    --colors=<NAME>
        Select color scheme from: "none", "dark". [default: "dark"]

    -l, --list
        Print selected column as list.

    --sep=<NAME>
        Set column separator. [default: " "]

    --long
        Print full information (each column is printed as a line).

    --debug=<FILE>
        Debug: read ``squeue -o "%all"`` from file.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseSLURM
"""
import argparse
import os
import pwd
import re
import sys

from .. import ps
from .. import rich
from .. import table
from .. import version


def main():

    # -- parse command line arguments --

    class Parser(argparse.ArgumentParser):
        def print_help(self):
            print(__doc__)

    parser = Parser()
    parser.add_argument("-U", action="store_true")
    parser.add_argument("-u", "--user", type=str, action="append", default=[])
    parser.add_argument("-p", "--pid", type=str, action="append")
    parser.add_argument("-c", "--command", type=str, action="append")
    parser.add_argument("-s", "--sort", type=str, action="append")
    parser.add_argument("-r", "--reverse", action="store_true")
    parser.add_argument("-o", "--output", type=str, action="append")
    parser.add_argument("--no-header", action="store_true")
    parser.add_argument("--no-truncate", action="store_true")
    parser.add_argument("--width", type=int)
    parser.add_argument("--colors", type=str, default="dark")
    parser.add_argument("-l", "--list", action="store_true")
    parser.add_argument("--sep", type=str, default=" ")
    parser.add_argument("--long", action="store_true")
    parser.add_argument("--debug", type=str)
    parser.add_argument("--version", action="version", version=version)
    args = vars(parser.parse_args())

    # -------------------------------- field-names and print settings --------

    # handle 'alias' options
    if args["U"]:
        args["user"] += [pwd.getpwuid(os.getuid())[0]]

    # conversion map: default field-names -> custom field-names
    alias = {
        "USER": "USER",
        "PID": "PID",
        "RSS": "MEM",
        "%CPU": "%CPU",
        "COMMAND": "COMMAND",
    }

    # conversion map: custom field-names -> default field-names
    aliasInv = {alias[key].upper(): key for key in alias}

    # rename command line options -> default field-names
    for key in [key for key in args]:
        if key.upper() in aliasInv:
            args[aliasInv[key.upper()]] = args.pop(key)

    # print settings of all columns
    # - "width"   : minimum width, adapted to print width (min_width <= width <= real_width)
    # - "align"   : alignment of the columns (except the header)
    # - "priority": priority of column expansing, columns marked "True" are expanded first
    columns = [
        {"key": "PID", "width": 3, "align": ">", "priority": True},
        {"key": "USER", "width": 7, "align": "<", "priority": True},
        {"key": "RSS", "width": 4, "align": ">", "priority": True},
        {"key": "%CPU", "width": 4, "align": ">", "priority": True},
        {"key": "COMMAND", "width": 10, "align": "<", "priority": True},
    ]

    # header
    header = {
        column["key"]: rich.String(alias[column["key"]], align=column["align"])
        for column in columns
    }

    # select color theme
    theme = ps.colors(args["colors"].lower())

    # -- load the output of "ps" --

    if not args["debug"]:

        lines = ps.read_interpret(theme=theme)

    else:

        lines = ps.read_interpret(
            data=open(args["debug"]).read(),
            theme=theme,
        )

    # ----------------------------- limit based on command-line options ------

    for key in ["USER", "PID", "COMMAND"]:

        if args[key]:

            # limit data
            lines = [
                line
                for line in lines
                if sum(1 if re.match(n, str(line[key])) else 0 for n in args[key])
            ]

            # color-highlight selected columns
            # - apply to all remaining lines
            for line in lines:
                line[key].color = theme["selection"]
            # - apply to the header
            header[key].color = theme["selection"]

    # -- sort --

    # default sort
    lines.sort(key=lambda line: line["RSS"], reverse=False)

    # optional: sort by key(s)
    if args["sort"]:

        for key in args["sort"]:

            lines.sort(key=lambda line: line[aliasInv[key.upper()]], reverse=args["reverse"])

    # -- select columns --

    if args["output"]:

        keys = [aliasInv[key.upper()] for key in args["output"]]

        columns = [column for column in columns if column["key"] in keys]

    # -- print --

    if True:

        # optional: print all fields and quit
        if args["long"]:

            table.print_long(lines)

            sys.exit(0)

        # optional: print as list and quit
        elif args["list"]:

            # - only one field can be selected
            if len(columns) > 1:
                print("Only one field can be selected")
                sys.exit(1)

            # - print and quit
            table.print_list(lines, columns[0]["key"], args["sep"])

            sys.exit(0)

        # default: print columns
        else:

            table.print_columns(
                lines,
                columns,
                header,
                args["no_truncate"],
                args["sep"],
                args["width"],
                not args["no_header"],
            )

            sys.exit(0)



import os
import subprocess
import unittest

import dummyslurm
import GooseSLURM


class Test_Gsub(unittest.TestCase):
    """
    Test Gsub commands.
    """

    def test_basic(self):

        myjob = "myjob.slurm"

        for filename in [myjob]:
            if os.path.isfile(filename):
                os.remove(filename)

        with open(myjob, "w") as file:
            file.write(GooseSLURM.scripts.plain(myjob))

        subprocess.check_output(["Gsub", "--quiet", myjob])

        os.remove(myjob)

    def test_logfile(self):

        myjob = "myjob.slurm"
        mylog = "mylog.yaml"

        for filename in [dummyslurm.logfile, myjob, mylog]:
            if os.path.isfile(filename):
                os.remove(filename)

        with open(myjob, "w") as file:
            file.write(GooseSLURM.scripts.plain(myjob))

        subprocess.check_output(["Gsub", "--quiet", "--log", mylog, myjob])

        log = GooseSLURM.fileio.YamlRead(mylog)

        self.assertIn(myjob, log)
        self.assertEqual(log[myjob], [1])

        subprocess.check_output(["Gsub", "--quiet", "--log", mylog, myjob])

        log = GooseSLURM.fileio.YamlRead(mylog)

        self.assertIn(myjob, log)
        self.assertEqual(log[myjob], [1, 2])

        os.remove(dummyslurm.logfile)
        os.remove(myjob)
        os.remove(mylog)

    def test_repeat(self):

        myjob = "myjob.slurm"

        for filename in [dummyslurm.logfile, myjob]:
            if os.path.isfile(filename):
                os.remove(filename)

        with open(myjob, "w") as file:
            file.write(GooseSLURM.scripts.plain(myjob))

        subprocess.check_output(["Gsub", "--quiet", "--repeat", "4", myjob])

        log = GooseSLURM.fileio.YamlRead(dummyslurm.logfile)

        for i, job in enumerate(log):

            if i:
                self.assertEqual(job["jobid"] - 1, job["dependency"])
            else:
                self.assertTrue("dependency" not in job)

        os.remove(dummyslurm.logfile)
        os.remove(myjob)

    def test_serial(self):

        myjobs = ["myjob_1.slurm", "myjob_2.slurm", "myjob_3.slurm"]

        for filename in [dummyslurm.logfile] + myjobs:
            if os.path.isfile(filename):
                os.remove(filename)

        for myjob in myjobs:
            with open(myjob, "w") as file:
                file.write(GooseSLURM.scripts.plain(myjob))

        subprocess.check_output(["Gsub", "--quiet", "--serial"] + myjobs)

        log = GooseSLURM.fileio.YamlRead(dummyslurm.logfile)

        for i, job in enumerate(log):

            if i:
                self.assertEqual(job["jobid"] - 1, job["dependency"])
            else:
                self.assertTrue("dependency" not in job)

        os.remove(dummyslurm.logfile)

        for myjob in myjobs:
            os.remove(myjob)


if __name__ == "__main__":

    unittest.main()



r"""Gdel
    Stop running jobs.

Usage:
    Gdel [options]
    Gdel [options] <JobId>...

Arguments:
    ID-number(s) of the job(s) to delete. (default: all user's jobs)

Options:

    ...
        Any option for ``Gstat``.

    -h, --help
        Show help.

    --version
        Show version.

(c - MIT) T.W.J. de Geus | tom@geus.me | www.geus.me | github.com/tdegeus/GooseSLURM
"""
import subprocess
import sys

import click

from . import Gstat


def main():

    cli_args = sys.argv[1:]
    gstat = Gstat.Gstat()
    gstat.parse_cli_args(cli_args)

    if len(gstat.args["jobs"]) == 0:
        cli_args += ["-U", "--status", "R", "--status", "PD"]
        gstat.parse_cli_args(cli_args)

    gstat.read()

    if len(gstat.lines) == 0:
        print("Nothing to do")
        return 0

    gstat.print()

    if not click.confirm("Delete above listed jobs?"):
        return 1

    cmd = ["scancel"] + [str(line["JOBID"]) for line in gstat.lines]

    if not gstat.args["debug"]:
        print(subprocess.run(cmd, capture_output=True, text=True).stdout, end="")
    else:
        print(cmd)



import re


def asSeconds(data, default=None):
    r"""
    Convert string to seconds. The following input is accepted:

    *   A humanly readable time (e.g. "1d").
    *   A SLURM time string (e.g. "1-00:00:00").
    *   A time string (e.g. "24:00:00").
    *   ``int`` or ``float``: interpreted as seconds.

    :arguments:

        **data** (``<str>`` | ``<float>`` | ``<int>``)
            The input string
            (number are equally accepted; they are directly interpreted as seconds).

    :options:

        **default** ([``None``] | ``<int>``)
            Value to return if the conversion fails.

    :returns:

        ``<int>``
            Number of seconds as integer (or default value if the conversion fails).
    """

    # convert int -> int (implicitly assume that the input is in seconds)
    if isinstance(data, int):
        return data

    # convert float -> float (implicitly assume that the input is in seconds)
    if isinstance(data, float):
        return int(data)

    # convert SLURM time string (e.g. "1-00:00:00")
    if re.match(r"^[0-9]*\-[0-9]*\:[0-9]*\:[0-9]*$", data):

        # - initialize number of days, hours, minutes, seconds
        t = [0, 0, 0, 0]
        # - split days
        if len(data.split("-")) > 1:
            t[0], data = data.split("-")
        # - split hours:minutes:seconds (all optional)
        data = data.split(":")
        # - fill from seconds -> minutes (if present) -> hours (if present)
        for i in range(len(data)):
            t[-1 * (i + 1)] = data[-1 * (i + 1)]
        # - return seconds
        return int(t[0]) * 24 * 60 * 60 + int(t[1]) * 60 * 60 + int(t[2]) * 60 + int(t[3])

    # convert time string in hours (e.g. "24:00:00")
    if re.match(r"^[0-9]*\:[0-9]*\:[0-9]*$", data):

        # - initialize number of hours, minutes, seconds
        t = [0, 0, 0]
        # - split hours:minutes:seconds (all optional)
        data = data.split(":")
        # - fill from seconds -> minutes (if present) -> hours (if present)
        for i in range(len(data)):
            t[-1 * i] = data[-1 * i]
        # - return seconds
        return int(t[0]) * 60 * 60 + int(t[1]) * 60 + int(t[2])

    # convert time string in minutes (e.g. "12:34")
    if re.match(r"^[0-9]*\:[0-9]*$", data):

        # - initialize number of minutes, seconds
        t = [0, 0]
        # - split hours:minutes:seconds (all optional)
        data = data.split(":")
        # - fill from seconds -> minutes (if present) -> hours (if present)
        for i in range(len(data)):
            t[-1 * i] = data[-1 * i]
        # - return seconds
        return int(t[0]) * 60 + int(t[1])

    # convert humanly readable time (e.g. "1d")
    if re.match(r"^[0-9]*\.?[0-9]*[a-zA-Z]$", data):

        if data[-1] == "d":
            return int(float(data[:-1]) * float(60 * 60 * 24))
        elif data[-1] == "h":
            return int(float(data[:-1]) * float(60 * 60))
        elif data[-1] == "m":
            return int(float(data[:-1]) * float(60))
        elif data[-1] == "s":
            return int(float(data[:-1]) * float(1))
        elif data[-1] == "w":
            return int(float(data[:-1]) * float(60 * 60 * 24 * 7))
        elif data[-1] == "M":
            return int(float(data[:-1]) * float(60 * 60 * 24 * 7 * 31))
        elif data[-1] == "y":
            return int(float(data[:-1]) * float(60 * 60 * 24 * 7 * 365))

    # one last try (implicitly assume that the input is in seconds)
    try:
        return int(data)
    except BaseException:
        pass

    # all conversions failed: return default value
    return default


def asUnit(data, unit, precision):
    r"""
    Convert to rich-string with a certain unit and precision. The output is e.g. ``"1.1d"``.

    :arguments:

        **data** (``<int>`` | ``<float>``)
            Numerical value (e.g. ``1.1``).

        **unit** (``<str>``)
            The unit (e.g. ``"d"``).

        **precision** (``<int>``)
            The precision with which to print (e.g. ``1``).

    :returns:

        ``<str>``
            The rich-string.
    """

    if precision:
        return f"{{0:.{precision:d}f}}{{1:s}}".format(data, unit)

    if abs(round(data)) < 10.0:
        return f"{data:.1f}{unit:s}"
    else:
        return f"{round(data):.0f}{unit:s}"


def asHuman(data, precision=None):
    r"""
    Convert to string that has the biggest possible unit.
    For example: ``100`` (seconds) -> ``"1.7m"``.

    :arguments:

        **data** (``<str>`` | ``<float>`` | ``<int>``)
            A time, see ``GooseSLURM.duration.asSeconds`` for conversion.

        **precision** (``<int>``)
            The precision with which to print. By default a precision of one is used for
            ``0 < value < 10``,
            while a precision of zero is used otherwise.

    :returns:

        ``<str>``
            The rich-string.
    """

    data = asSeconds(data)

    if data is None:
        return ""

    units = (
        (24 * 60 * 60, "d"),
        (60 * 60, "h"),
        (60, "m"),
        (1, "s"),
    )

    for val, unit in units:
        if abs(data) >= val:
            return asUnit(float(data) / float(val), unit, precision)

    return asUnit(float(data), "s", precision)


def asSlurm(data):
    r"""
    Convert to a SLURM time string. For example ``"1d"`` -> ``"1-00:00:00"``.

    :arguments:

        **data** (``<str>`` | ``<float>`` | ``<int>``)
            A time, see ``GooseSLURM.duration.asSeconds`` for conversion.

    :returns:

        ``<str>``
            The rich-string.
    """

    data = asSeconds(data)

    if data is None:
        return ""

    s = int(data % 60)
    data = (data - s) / 60
    m = int(data % 60)
    data = (data - m) / 60
    h = int(data % 24)
    data = (data - h) / 24
    d = int(data)

    return "%d-%02d:%02d:%02d" % (d, h, m, s)



import re


def asBytes(data, default=None, default_unit=1):
    r"""
    Convert string to bytes. The following input is accepted:

    *   A humanly readable string (e.g. "1G").
    *   ``int`` or ``float``: interpreted as bytes.

    :arguments:

        **data** (``<str>`` | ``<float>`` | ``<int>``)
            The input string (number are equally accepted; they are directly interpreted as bytes).

    :options:

        **default** ([``None``] | ``<int>``)
            Value to return if the conversion fails.

        **default_unit** (``int``)
            The unit to assume if no unit if specified (specify the number of bytes).

    :returns:

        ``<int>``
            Number of bytes as integer (or default value if the conversion fails).
    """

    # convert int -> int (assume that the unit)
    if isinstance(data, int):
        return data * int(default_unit)

    # convert float -> float (assume that the unit)
    if isinstance(data, float):
        return int(data * default_unit)

    # convert humanly readable time (e.g. "1G")
    if re.match(r"^[0-9]*\.?[0-9]*[a-zA-Z]$", data):

        if data[-1] == "T":
            return int(float(data[:-1]) * 1.0e12)
        if data[-1] == "G":
            return int(float(data[:-1]) * 1.0e9)
        if data[-1] == "M":
            return int(float(data[:-1]) * 1.0e6)
        if data[-1] == "K":
            return int(float(data[:-1]) * 1.0e3)

    # one last try (assume that the unit)
    try:
        return int(float(data) * default_unit)
    except BaseException:
        pass

    # all conversions failed: return default value
    return default


def asUnit(data, unit, precision):
    r"""
    Convert to rich-string with a certain unit and precision. The output is e.g. ``"1.1G"``.

    :arguments:

        **data** (``<int>`` | ``<float>``)
            Numerical value (e.g. ``1.1``).

        **unit** (``<str>``)
            The unit (e.g. ``"G"``).

        **precision** (``<int>``)
            The precision with which to print (e.g. ``1``).

    :returns:

        ``<str>``
            The rich-string.
    """

    if precision:
        return f"{{0:.{precision:d}f}}{{1:s}}".format(data, unit)

    if abs(round(data)) < 10.0:
        return f"{data:.1f}{unit:s}"
    else:
        return f"{round(data):.0f}{unit:s}"


def asHuman(data, precision=None):
    r"""
    Convert to string that has the biggest possible unit. For example ``1e6`` (bytes) -> ``"1.0M"``.

    :arguments:

        **data** (``<str>`` | ``<float>`` | ``<int>``)
            An amount of memory, see ``GooseSLURM.duration.asBytes`` for conversion.

        **precision** (``<int>``)
            The precision with which to print. By default a precision of one is used for
            ``0 < value < 10``,
            while a precision of zero is used otherwise.

    :returns:

        ``<str>``
            The rich-string.
    """

    data = asBytes(data)

    if data is None:
        return ""

    units = (
        (1e12, "T"),
        (1e9, "G"),
        (1e6, "M"),
        (1e3, "K"),
        (1, "B"),
    )

    for val, unit in units:
        if abs(data) >= val:
            return asUnit(float(data) / float(val), unit, precision)

    return asUnit(float(data), "B", precision)


def asSlurm(data):
    r"""
    Convert to a SLURM string. For example ``"1G"``.

    :arguments:

        **data** (``<str>`` | ``<float>`` | ``<int>``)
            An amount of memory, see ``GooseSLURM.duration.asBytes`` for conversion.

    :returns:

        ``<str>``
            The rich-string.
    """

    data = asBytes(data)

    if data is None:
        return ""

    if data % 1e12 == 0:
        return str(int(data / 1e12)) + "T"
    if data % 1e9 == 0:
        return str(int(data / 1e9)) + "G"
    if data % 1e6 == 0:
        return str(int(data / 1e6)) + "M"
    if data % 1e3 == 0:
        return str(int(data / 1e3)) + "K"

    return str(data)



import GooseSLURM as gs

# job-options
sbatch = {
    "job-name": "json",
    "out": "job.slurm.out",
    "nodes": 1,
    "ntasks": 1,
    "cpus-per-task": 1,
    "partition": "debug",
}

# command to run
command = """
# Write job info to a JSON-log file
# =================================

# get hostname
myhost=`hostname`

# write log file
cat <<EOF > job.slurm.json
{
  "hostname"            : "${myhost}",
  "SLURM_SUBMIT_DIR"    : "${SLURM_SUBMIT_DIR}",
  "SLURM_JOB_ID"        : "${SLURM_JOB_ID}",
  "SLURM_JOB_NODELIST"  : "${SLURM_JOB_NODELIST}",
  "SLURM_SUBMIT_HOST"   : "${SLURM_SUBMIT_HOST}",
  "SLURM_JOB_NUM_NODES" : "${SLURM_JOB_NUM_NODES}",
  "SLURM_CPUS_PER_TASK" : "${SLURM_CPUS_PER_TASK}"
}
EOF

# Execute
# =======

# simplest example in the world, sleep a bit to allow a bit of monitoring
echo "hello world" > "test.log"
sleep 10
"""

# write SLURM file
open("job.slurm", "w").write(gs.scripts.plain(command=command, **sbatch))



import functools
import operator
import os

import click
import yaml


def ContinueDump(filename):
    r"""
    Check to proceed dumping a file
    (checks to create the directory and/or to overwrite an existing file).
    Return ``True`` is the user confirms to proceed.
    """

    dirname = os.path.dirname(filename)

    if os.path.isfile(filename):
        if not click.confirm(f'Overwrite "{filename:s}"?'):
            return False

    elif not os.path.isdir(dirname) and len(dirname) > 0:
        if not click.confirm(f'Create "{os.path.dirname(filename):s}"?'):
            return False

    return True


def YamlRead(filename):
    r"""
    Read YAML file and return its content as ``list`` or ``dict``.
    """

    if not os.path.isfile(filename):
        raise OSError(f'"{filename:s} does not exist')

    with open(filename) as file:
        return yaml.load(file.read(), Loader=yaml.FullLoader)


def YamlGetItem(filename, key=[]):
    r"""
    Get an item from a YAML file.
    Optionally the key to the item can be specified as a list. E.g.
    *   ``[]`` for a YAML file containing only a list.
    *   ``['foo']`` for a plain YAML file.
    *   ``['key', 'to', foo']`` for a YAML file with nested items.
    """

    data = YamlRead(filename)

    if len(key) == 0 and not isinstance(data, list):
        raise OSError(f'Specify key for "{filename:s}"')

    if len(key) > 0:
        try:
            return functools.reduce(operator.getitem, key, data)
        except BaseException:
            raise OSError('"{:s}" not in "{:s}"'.format("/".join(key), filename))

    return data


def YamlDump(filename, data):
    r"""
    Dump data (as ``list`` or ``dict``) to YAML file.
    Unless ``force = True`` the function prompts before overwriting an existing file.
    """

    dirname = os.path.dirname(filename)

    if not os.path.isdir(dirname) and len(dirname) > 0:
        os.makedirs(os.path.dirname(filename))

    with open(filename, "w") as file:
        yaml.dump(data, file)



from . import rich


def print_long(lines):
    r"""
    Print full data without much formatting. The output looks as follows:

    .. code-block:: none

      +------------------+
      | line 1, column 1 |
      | line 1, column 2 |
      | ...              |
      +-----------------+
      | line 2, column 1 |
      | line 2, column 2 |
      | ...              |
      +------------------+

    :arguments:

      **lines** (``[ {'JOBID': '1234', ...}, ...]``)
        List of lines, with each line stored as a dictionary.
        Note that all data has to be stored as one as string,
        or as one the GooseSLURM.rich classes (no rich printing is used though).
    """

    # width of the field-names
    # - initialize
    width = 0
    # - compute
    for line in lines:
        for key in sorted(line):
            width = max(width, len(key))

    # print format
    head = "{{key:<{width:d}.{width:d}s}}".format(width=width + 1)
    fmt = "{{key:<{width:d}.{width:d}s}}: {{data:s}}".format(width=width + 1)

    # print header
    print(head.format(key="-" * 100, data=""))

    # print data
    for line in lines:
        for key in sorted(line):
            print(fmt.format(key=key, data=str(line[key])))
        print(head.format(key="-" * 100, data=""))


def print_columns(
    lines, columns, header, no_truncate=False, sep=", ", cols=None, print_header=True
):
    r"""
    Print table to fit the screen. This function can show data truncated, or even suppress columns
    if there is insufficient room.

    :arguments:

      **lines** (``[ {'JOBID': '1234', ...}, ...]``)
        List of lines, with each line stored as a dictionary.
        Note that all data has to be stored as one of the GooseSLURM.rich classes
        (to customize the color, precision, ...) or as string.

      **columns** (``[ {'key': 'JOBID', 'width': 7, 'align': '>', 'priority': True}, ...]``)
        List with print settings of each column:
        - 'key'     : the key-name used to store each line (see ``lines`` below)
        - 'width'   : minimum print width (expanded as much as possible to fit the data)
        - 'align'   : alignment of the column
        - 'priority': priority of column expansion, columns marked ``True`` are expanded first

      **header** (``{'JOBID': 'JobID', ...}``)
        Header name for each column.

    :options:

      **no_truncate** ([``False``] | ``True``)
        Disable truncation of columns. In this case each column is expanded to fit the data.

      **sep** ([``', '``] | ``<str>``)
        Separator between columns.

      **cols** ([``None``] | ``<int>``)
        Number of characters on one line. If ``None`` the current terminal's width is used.

      **print_header** ([``True``] | ``False``)
        Optionally skip printing of header.
    """

    # check available data
    # --------------------

    # function to check if a "key" present in all lines
    def inlines(lines, key):
        # - check all lines
        for line in lines:
            if key not in line:
                return False
        # - all lines passed: return True
        return True

    # select columns based on data availability
    columns = [column for column in columns if inlines(lines, column["key"])]

    # select header based on columns
    header = {column["key"]: header[column["key"]] for column in columns}

    # convert to GooseSLURM.rich
    # --------------------------

    for line in lines:
        for key in line:
            if not isinstance(line[key], rich.String):
                line[key] = rich.String(line[key])

    for key in header:
        if not isinstance(header[key], rich.String):
            header[key] = rich.String(header[key])

    # compute column-width, based on data
    # -----------------------------------

    # initialize
    for column in columns:
        column["real"] = 0

    # get actual width
    # - header
    for column in columns:
        column["real"] = max(column["real"], len(str(header[column["key"]])))
    # - data
    for line in lines:
        for column in columns:
            column["real"] = max(column["real"], len(str(line[column["key"]])))

    # auto-limit columns, auto-adjust their width
    # -------------------------------------------

    if no_truncate:

        # set all columns to the real width
        for line in lines:
            for column in columns:
                column["width"] = column["real"]

    else:

        # get the terminal size
        if cols is None:
            import shutil

            cols, _ = shutil.get_terminal_size()

        # get the cumulative minimum size of the columns (+ spacing between the columns)
        # - first entry
        columns[0]["total"] = columns[0]["width"]
        # - all other entries
        for i in range(1, len(columns)):
            columns[i]["total"] = columns[i - 1]["total"] + columns[i]["width"] + len(sep)

        # truncate at terminal size
        columns = [column for column in columns if column["total"] <= cols]

        # get the available size to expand
        room = cols - columns[-1]["total"]

        # expand minimum width, as long there is room
        # - distinguish priorities for expanding
        low = [column["key"] for column in columns if not column.get("priority", False)]
        high = [column["key"] for column in columns if column["key"] not in low]
        # - expand
        for prio in [high, low]:
            for column in columns:
                if column["key"] not in prio:
                    continue
                if room <= 0:
                    break
                dw = min(column["real"] - column["width"], room)
                if dw <= 0:
                    continue
                column["width"] += dw
                room -= dw

    # apply width
    for line in lines:
        for column in columns:
            line[column["key"]].width = column["width"]
            line[column["key"]].align = column["align"]

    # print to screen
    # ---------------

    # select header based on columns
    header = {column["key"]: header[column["key"]] for column in columns}

    # apply width to header
    for column in columns:
        header[column["key"]].width = column["width"]

    # separator
    # - copy from header
    hline = {key: rich.String(**value.__dict__) for key, value in header.items()}
    # - convert from header (print format retained)
    for key in hline:
        hline[key].data = "=" * hline[key].width

    # header
    if print_header:
        print(sep.join(hline[column["key"]].format() for column in columns))
        print(sep.join(header[column["key"]].format() for column in columns))
        print(sep.join(hline[column["key"]].format() for column in columns))
    # data
    for line in lines:
        print(sep.join(line[column["key"]].format() for column in columns))


def print_list(lines, key, sep=" "):
    r"""
    Print a single column as a list.

    :arguments:

      **lines** (``[ {'JOBID': '1234', ...}, ...]``)
        List of lines, with each line stored as a dictionary.
        Note that all data has to be stored as one as string,
        or as one the GooseSLURM.rich classes (no rich printing is used though).

      **key** (``'JOBID'``)
        Column to print.

    :options:

      **sep** ([``' '``] | ``<str>``)
        Separator between columns.
    """

    for line in lines:
        print(line[key], end=sep)

    print("")



from pathlib import Path

from setuptools import find_packages
from setuptools import setup

project_name = "GooseSLURM"

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name=project_name,
    license="MIT",
    author="Tom de Geus",
    author_email="tom@geus.me",
    description="Support functions for SLURM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="SLURM",
    url=f"https://github.com/tdegeus/{project_name:s}",
    packages=find_packages(exclude=["tests"]),
    use_scm_version={"write_to": f"{project_name}/_version.py"},
    setup_requires=["setuptools_scm"],
    install_requires=["click", "tqdm", "PyYAML", "numpy"],
    entry_points={
        "console_scripts": [
            f"Gdel = {project_name}.cli.Gdel:main",
            f"Ginfo = {project_name}.cli.Ginfo:main",
            f"Gps = {project_name}.cli.Gps:main",
            f"Gstat = {project_name}.cli.Gstat:main",
            f"Gsub = {project_name}.cli.Gsub:main",
            f"Gacct = {project_name}.sacct:_Gacct_catch",
        ]
    },
)
